/**
 * Example of how to call the TCB DistributedMeasurement service.
 * 
 * This client demo is responsible for starting a distributed measurement
 * that is ended by server demo. Only server demo presents the measurement
 * results.
 * 
 * @author Paulo Sousa
 */

#include "stdafx.h"
//include sockets header file
#include <winsock.h>
//include tcb services library header file
#include "..\..\..\..\tcb_services_library\tcb_services_library.h"


//TODO: replace this values according to your situation
#define MY_IP "192.168.0.1"
#define SERVER_IP "192.168.0.2"
#define SERVER_PORT 40000


int main(int argc, char* argv[]) {

	//open TCB session
	printf("Opening TCB session\n");
	int error = open_tcb_session();
	if (error!=0) {
		printf("Error opening session=%d\n",error);
		getchar();
		return -1;
	}

	//create an UDP socket
	printf("Creating socket\n");
	SOCKET s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s==INVALID_SOCKET) {
		printf("Error creating socket: %d\n", error);
		getchar();
		return -1;
	}

	//bind socket
	printf("Binding socket\n");
	struct sockaddr_in sa;
	sa.sin_family=AF_INET;
	sa.sin_addr.s_addr= inet_addr(MY_IP);
	sa.sin_port=htons(0);
	error= bind(s, (struct sockaddr*)&sa, sizeof(sa));
	if (error!=0) {
		printf("Error binding socket: %d\n", error);
		getchar();
		return -1;
	}

	//start a distributed measurement
	printf("Starting a dist. measurement\n");
	tcb_dist_action_id_t action_id;
	error=start_tcb_dist_measurement(get_tcb_timestamp(), &action_id);
	if (error!=0) {
		printf("Error starting dist measurement:");
		switch (error) {
			case -1: printf("no TCB session opened!\n");
					 break;
			default: printf("unknown\n");
		}
		getchar();
		return -1;
	}
	
	//send a msg and serialize the identifier of the measurement in it
	char buf[50+sizeof(action_id)];
	//action_id goes in first sizeof(action_id) bytes
	memcpy(buf, (char*)&action_id, sizeof(action_id));
	strcpy(buf+sizeof(action_id), "Hi server!");
	struct sockaddr_in to;
	to.sin_family = AF_INET;
	to.sin_addr.s_addr = inet_addr(SERVER_IP);
	to.sin_port = htons(SERVER_PORT);
		
	int nbytes = sendto(s, buf, 50+sizeof(action_id), NULL, (struct sockaddr*)&to, 
						    sizeof(to));
		
	if (nbytes<0) {
		printf("Error sending msg: %d\n", nbytes);
	}
	else printf("%d bytes sent\n", nbytes);

	//close socket
	printf("Closing socket\n");
	error = closesocket(s);
	if (error!=0) {
		printf("Error closing socket=%d\n",error);
		getchar();
		return -1;
	}

	//close TCB session
	printf("Closing TCB session\n");
	error=close_tcb_session();
	if (error!=0) {
		printf("Error closing session=%d\n",error);
		getchar();
		return -1;
	}

	//wait for user to press a key
	printf("Press a key to close window...\n");
	getchar();
	return 0;
}
