/**
 * Example of how to call the TCB LocalFailureDetection service.
 * 
 * To call this service, you need first to create a DLL with
 * two functions:
 * 
 * 		1) The function to timely execute if a failure occurs
 * 		2) A function that returns the Worst Case Execution Time
 * 		   (milliseconds) of the function specified in 1). 
 * 
 * There is already a dll example created that works with this example.
 * The source code (Embedded Visual C++ 3.0) is located in directory
 * "tcb_services_examples/dll_example_for_tcb_handler". You should compile it
 * and deploy in PocketPC prior to running this example.
 * 
 * 
 * @author Paulo Sousa
 */

#include "stdafx.h"
//include tcb services library header file
#include "..\..\..\tcb_services_library\tcb_services_library.h"

int main(int argc, char* argv[]){

	//open_session
	printf("Opening TCB session\n");
	int error = open_tcb_session();
	if (error!=0) {
		printf("Error opening session=%d\n",error);
		getchar();
		return -1;
	}


	//start_local_fd
	printf("Starting local fd\n");
	//time to failure (ms)
	int spec = 15;
	printf("Specification: %d ms\n", spec);
	//maximum time allowed for the execution of failure handler (ms)
	int deadline = 0;
	char dllName[100];
	char funcName[100];
	char wcetName[100];
	//DLL where resides the function and the wcet function
	strcpy(dllName, "tcb_handler_example.dll");
	//function name
	strcpy(funcName, "localFD_example");
	//wcet function name
	strcpy(wcetName, "localFD_example_wcet");
	//get an initial instant
	int start_ev = get_tcb_timestamp();
	int tag = start_tcb_local_fd(start_ev, spec, deadline, 
								 dllName, funcName, wcetName);
	if (tag<0) {
		printf("Error starting local FD: %d\n", tag);
	}
	else {
		//end_local_fd
		tcb_local_fd_info_t lfi =  end_tcb_local_fd(tag);
		printf("Duration: %d ms\n", lfi.procdelay);
		//verify if a timing failure occured
		if (lfi.faulty) 
			printf("Timing failure!\n");
		else
			printf("Timely!\n");
	}


	//close_session
	printf("Closing session\n");
	error=close_tcb_session();
	if (error!=0) {
		printf("Error closing session=%d\n",error);
		getchar();
		return -1;
	}
	printf("Close ok!\n");

	//wait for user to press a key
	printf("Press a key to close window...\n");
	getchar();
	return 0;
}
