/**
 * Example of how to call the TCB TimelyExecution service.
 * 
 * To call this service, you need first to create a DLL with
 * two functions:
 * 
 * 		1) The function to timely execute
 * 		2) A function that returns the Worst Case Execution Time
 * 		   (milliseconds) of the function specified in 1). 
 * 
 * There is already a dll example created that works with this example.
 * The source code (Embedded Visual C++ 3.0) is located in directory
 * "tcb_services_examples/dll_example_for_tcb_handler". You should compile it
 * and deploy in PocketPC prior to running this example.
 * 
 * 
 * @author Paulo Sousa
 */

#include "stdafx.h"
//include tcb services library header file
#include "..\..\..\tcb_services_library\tcb_services_library.h"

int main(int argc, char* argv[]){
	
	//open_session
	printf("Opening TCB session\n");
	int error = open_tcb_session();
	if (error!=0) {
		printf("Error opening session=%d\n",error);
		getchar();
		return -1;
	}

	//start timely_exec
	printf("Starting timely exec\n");
	char dllName[100];
	char funcName[100];
	char wcetName[100];
	//delay for a deferred execution (ms)
	int delay=50;
	//maximum time allowed for the execution (ms) [includes delay]
	int max_exec = 100;
	printf("Timely exec starting with\ndelay= %d ms and max_exec= %d ms\n", delay, max_exec);
	//DLL where resides the function and the wcet function
	strcpy(dllName, "tcb_handler_example.dll");
	//function name
	strcpy(funcName, "timely_example");
	//wcet function name
	strcpy(wcetName, "timely_example_wcet");
	//get an initial instant
	int start_ev = get_tcb_timestamp();
	int end_ev=tcb_timely_exec(start_ev, delay, max_exec, dllName, funcName, wcetName);
	if (end_ev<0) {
		printf("Error in timely exec: %d\n", end_ev);
	}
	else {
		//Note: exec_time includes delay
		printf("Timely exec OK!, exec_time=%d\n", end_ev-start_ev);
	}

	//close_session
	printf("Closing session\n");
	error=close_tcb_session();
	if (error!=0) {
		printf("Error closing session=%d\n",error);
		getchar();
		return -1;
	}
	printf("Close ok!\n");

	//wait for user to press a key
	printf("Press a key to close window...\n");
	getchar();
	return 0;
}
