/**
 * TCB for Windows CE 3.0
 *
 * This file contains all the source code.
 *
 * @author Paulo Sousa
 * (based on RTAI TCB implementation by Pedro Martins)
 * @date 18/Nov/2002
 */

#define UNICODE



//header file with the standard includes
#include "stdafx.h"
//header file with the TCB configuration parameters
#include "config_parameters.h"
//header file with the TCB structures
#include "structures.h"
//header file with socket functions
#include "Winsock.h"
//header file with string functions
#include <string.h>

#define MAJOR_VERSION 0
#define MINOR_VERSION 2
#define TINY_VERSION 1


//turns on/off network access protocol debug mode
//#define debug_net(x) printf x  //ON
#define debug_net(x)			 //OFF

//another network debug: more restrict
//#define debug_net1(x) printf x			 //ON
#define debug_net1(x) 			 //OFF

//turns on/off TCB debug mode
//#define debug_tcb(x) printf x	//ON
#define debug_tcb(x)			//OFF

//turns on/off TCB services debug mode
//#define debug_tcb_services(x) printf x	//ON
#define debug_tcb_services(x)			//OFF

//turns on/off another TCB services debug mode: more restrict
//#define debug_tcb_services1(x) printf x	//ON
#define debug_tcb_services1(x)			//OFF

//socket for communication
SOCKET s = NULL;

//stores our id
unsigned long my_id;

//stores information about the last data msg received from other machines
//msg_info_type last_rcv_info[NUMBER_OF_MACHINES];
//my data sequence number 
int data_seq_num=0;

//send task (sends tcb messages)
HANDLE hSend = NULL;
//create receive task (receives tcb messages)
HANDLE hRecv = NULL;
//NOTE: this tasks are declared here because they
//need to be accessed from different functions. Tasks without this requirement
//are declared in the main() function.

//total number of omissions detected
int n_omissions=0;

//event used to TCB DLL sinalize ServicesThread that a request is ready
HANDLE tcb_wait_event=NULL;
//event used to ServicesThread sinalize TCB DLL that an answer is ready
HANDLE dll_wait_event=NULL;
//event used to TCB DLL wait for TCB distributed timing failure detection 
//service info
HANDLE wait_info_event=NULL;

//used for communication TCB DLL<->ServicesThread through shared memory
service_request_response_t *srrt=NULL;

//used for communication DistFDService->TCB DLL waitInfo() through shared memory
fail_info_to_app_t *fi=NULL;

//used by the local measurement service
int localMeasures[MAX_LOCAL_MEASURES];

//used by the timely execution service
exec_info_t	executions[MAX_EXEC];

//used by the local timing failure detection service
local_fd_info_t localFD[MAX_LOCAL_FD];

//buffer to auxiliate receives
//char buffer[sizeof(data_msg_t)];

//to synchronize accesses to TCB tables
CRITICAL_SECTION tcb_critical_section;

//to synchronize accesses to loopback buffer
CRITICAL_SECTION loopback_critical_section;
/*
//TODO: por a struct e o define no sitio certo
typedef struct loopback_buffer {
	char buffer[sizeof(data_msg_t)];
	int size;
} loopback_buffer_t;
#define MAX_SIZE_LOOPBACK_BUFFER 10
loopback_buffer_t loopbt[MAX_SIZE_LOOPBACK_BUFFER];
int loop_first=-1, loop_last=-1;
*/

//stores the number of distributed FD running with each machine
//int number_of_detections[NUMBER_OF_MACHINES];

//stores the number of distributed FD services started in the current
//send period
int n_sends_in_this_slot;

//stores the distributed FD requests that were not yet disseminated
//in the control channel
send_info_t send_info[MAX_SEND_INFO_TABLE_SIZE];

//stores the distributed FD requests received in the control channel,
//that haven't yet finished 
rcv_info_t receive_info[MAX_RCV_INFO_TABLE_SIZE];

//stores information about msgs received in the payload channel and
//whose failure info was not yet received by TCB in the control channel
rcv_payload_info_t receive_payload_info[MAX_RCV_PLOAD_TABLE_SIZE];

//seq number used to identify distributed FD requests
int tcb_seq_num=1;

//stores the current state of distributed FD operations until
//they finish and are sent to application
wait_info_t wait_info[MAX_WAIT_INFO_TABLE_SIZE];

//for passing arguments to DistFDThread
dist_fd_info_t dfi;

//stores the remote FD handler. There exists only one because we only
//support 1 simulatneous application using the TCB.
distfunctype remote_fd_handler = NULL;
int remote_fd_deadline;

//stores information about the monitoring of event channels QoS
qos_monitor_info_t qos_monitor_info[MAX_QOS_MONITOR_INFO_TABLE_SIZE];

/* Global Clock */
//info about the global clock of each machine
//global_clock_info_t gclock_info[NUMBER_OF_MACHINES];
//value of global clock on synchronization instant
int vclock_at_sync;
//synchronization instant
int time_of_sync;
//drift of global clock
double virtual_drift;
//drift validity
int apply_drift_until;

//stores the instant when TCB started
int TCBstartTime;

//stores the Event Channels ids to which each machine is subscribed
//int ec_ids[NUMBER_OF_MACHINES];

//stores the TCBs currently running (and in reach)
tcb_info_t membership[MAX_NUMBER_OF_MACHINES];

#ifdef TCB_MEASUREMENTS

//stores the sum of all roundtrip values from each machine
//int total_roundtripVal[MAX_NUMBER_OF_MACHINES];
//stores the number of roundtrip values recved from each machine
//int num_roundtripVal[MAX_NUMBER_OF_MACHINES];
//stores the minimum roundtrip value from each machine
//int min_roundtripVal[MAX_NUMBER_OF_MACHINES];
//stores the maximum roundtrip value from each machine
//int max_roundtripVal[MAX_NUMBER_OF_MACHINES];

//stores the sum of all token delays
int total_token_delay=0;
//stores the number of token delays received
int num_token_delay=0;
//stores the minimum token delay
int min_token_delay=100000;
//stores the maximum token delay
int max_token_delay=0;
//stores instant when the last token was received
int last_token_instant = -1;

//stores the time that a msg waits in the send_queue
int max_in_send_queue_latency=0;
int min_in_send_queue_latency=100000;
int num_in_send_queue_latency=0;
int total_in_send_queue_latency=0;

//stores the maximum time since a msg is received until being pushed
//into the receive queue
int max_before_recv_queue_latency=0;

//stores the number of times that the receive_queue exceed its maximum
//expected size (=NUMBER_OF_MACHINES)
int num_rcv_queue_overflow=0;


#endif



/************************************************************
 * Auxiliary functions
 ************************************************************/

/**
 * FUNCTION addTCB
 *
 * GOAL: To add a TCB to membership
 *
 * PARAM: tcb_id Id of the TCB to add
 *
 * RETURN: 0 if OK, -1 if TCB already exists, -2 if full
 */
 int addTCB(unsigned long tcb_id){

	 EnterCriticalSection(&tcb_critical_section);

	 int free_pos=-1;
	 for (int i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		//verify if it already exists
		if (!membership[i].empty && membership[i].id==tcb_id) {
			LeaveCriticalSection(&tcb_critical_section);
			return -1;
		}
		else {
			//store first free position
			if (membership[i].empty && free_pos==-1)
				free_pos=i;
		}
	 }
	 if (free_pos==-1) {
		 LeaveCriticalSection(&tcb_critical_section);
		 return -2;
	 }

	 //add TCB
	 membership[free_pos].empty = FALSE;
	 membership[free_pos].id = tcb_id;
	 membership[free_pos].last_rcv_info.seq = -1;
	 membership[free_pos].last_rcv_info.snd_time = 0;
	 membership[free_pos].last_rcv_info.timestamp = 0;
	 membership[free_pos].number_of_detections = 0;
	 membership[free_pos].gclock_info.error = 0;
	 membership[free_pos].gclock_info.time_of_sync = 0;
	 membership[free_pos].ec_ids = 0;
	 membership[free_pos].first_seq = 0;
	 membership[free_pos].arrival = GetTickCount();

#ifdef TCB_MEASUREMENTS

	membership[free_pos].max_roundtripVal = 0;
	membership[free_pos].min_roundtripVal = 10000;
	membership[free_pos].num_roundtripVal = 0;
	membership[free_pos].total_roundtripVal = 0;

#endif

	 LeaveCriticalSection(&tcb_critical_section);

	 return 0;
 }


/**
 * FUNCTION deleteTCB
 *
 * GOAL: To delete a TCB from membership
 *
 * PARAM: tcb_id Id of the TCB to delete
 *
 * RETURN: 0 if OK, -1 if TCB doesn't exist
 */
 int deleteTCB(unsigned long tcb_id){
	 bool found = FALSE;
	 EnterCriticalSection(&tcb_critical_section);
		
	 for (int i=0; i<MAX_NUMBER_OF_MACHINES && !found; i++) {
		//verify if it exists
		if (!membership[i].empty && membership[i].id==tcb_id) {
			found = TRUE;
			membership[i].empty = TRUE;
		}
	 }
	 //if it doesn't exist
	 if (!found) {
		 LeaveCriticalSection(&tcb_critical_section);
		 return -1;
	 }

	 LeaveCriticalSection(&tcb_critical_section);
	 return 0;
 }



 /**
 * FUNCTION getTCB
 *
 * GOAL: To get a TCB from membership
 *
 * PARAM: tcb_id Id of the TCB to get
 *
 * RETURN: TCB info if OK, NULL if TCB doesn't exist
 */
 tcb_info_t* getTCB(unsigned long tcb_id){

	 EnterCriticalSection(&tcb_critical_section);

	 for (int i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		if (!membership[i].empty && membership[i].id==tcb_id) {
			LeaveCriticalSection(&tcb_critical_section);
			return &membership[i];
		}
	 }

	 LeaveCriticalSection(&tcb_critical_section);
	 return NULL;
 }


/**
 * FUNCTION: wait_user
 *
 * GOAL: to wait for a user input
 */
void wait_user () {
	printf("Press any key to continue\n");
	char c;
	scanf("%c", &c);
}



/**********************************************************
 * Global clock functions
 **********************************************************/


/**
 * FUNCTION get_vclock_at
 *
 * GOAL: to return the global clock value on the (local clock) instant specified.
 * 
 * PARAM: 
 *		time - local clock instant (<=GetTickCount())
 *
 * RETURNS: the global clock value on the (local clock) instant specified.
 */
int get_vclock_at(int time){	
	int elapsed;	

	if (virtual_drift == 0){
		elapsed = time - time_of_sync;	
		return vclock_at_sync + elapsed;
	}
	
	if (time > apply_drift_until){
		elapsed = time - apply_drift_until;	
		vclock_at_sync = (int)(vclock_at_sync + ((double)(apply_drift_until-time_of_sync)*virtual_drift) + (apply_drift_until - time_of_sync)); 	
		virtual_drift = 0;		
		time_of_sync = apply_drift_until;
		return vclock_at_sync + elapsed;	
	}	
	elapsed = time - time_of_sync;	

	return (int)((vclock_at_sync + elapsed + ((double)elapsed * virtual_drift)));
}


/**
 * FUNCTION get_vclock
 *
 * GOAL: to return the present global clock value.
 * 
 * RETURNS: the present global clock value.
 */
int get_vclock(void){
	return get_vclock_at(GetTickCount());
}


/**
 * FUNCTION init_vclock
 *
 * GOAL: to init the global clock protocol.
 */
void init_vclock(void){
	int i,t;

	//initialize protocol variables only when no msg of other TCBs
	//was received
	for (i=0;i<MAX_NUMBER_OF_MACHINES;i++){
		if (!membership[i].empty && membership[i].gclock_info.vclock!=0){
			break;
		}
	}
	if (i == MAX_NUMBER_OF_MACHINES){
		virtual_drift = 0;	
		t = GetTickCount();
		vclock_at_sync = t;
		time_of_sync = t;
	}
}


/**
 * FUNCTION update_vclock
 *
 * GOAL: to update the global clock
 *
 * @param ti Info of sender TCB
 * @param from_vclock Msg sender vclock value
 * @param rndtrip Msg delivery delay
 * @param rcv_time Msg rcv_time
 * @param init_state Indicates if msg was sent in init_state
 */
void update_vclock(tcb_info_t *ti, int from_vclock, int rndtrip, int rcv_time, bool init_state){
	int i;
	double curr_error = 0;

	//if msg was not sent by me
	if (ti->id != my_id){
		//if msg brings rndtrip info
		if (rndtrip!=-1){
			//calculate error since last update
			curr_error = ti->gclock_info.error + ((double)(rcv_time-ti->gclock_info.time_of_sync)*2*CLOCK_DRIFT_RATE);
			//if error is greater than delivery delay or it's the first msg received from sender
			if ((curr_error > ((rndtrip >> 1))) || (ti->gclock_info.time_of_sync==0)) {
				ti->gclock_info.vclock = from_vclock + ((rndtrip >> 1));	
				ti->gclock_info.error = ((rndtrip >> 1));
				ti->gclock_info.time_of_sync = rcv_time;
			}
			else {
				return ;
			}
		}
		//if not, do calculations with MAXIMUM_DELIVERY_DELAY
		else {
			ti->gclock_info.vclock = from_vclock + ((MAXIMUM_DELIVERY_DELAY >> 1));	
			ti->gclock_info.error = ((MAXIMUM_DELIVERY_DELAY >> 1));
			ti->gclock_info.time_of_sync = rcv_time;
			return ;
		}
	}
	
	
	//find the machine with the smallest id that has its clock initialized
	unsigned long min_id = 0;
	for (i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		if (!membership[i].empty) {
			//clock initialized
			if (membership[i].gclock_info.vclock!=0) {
				if (membership[i].id<min_id || min_id==0) {
					min_id = membership[i].id;
				}
			}

		}
	}
	tcb_info_t* min_ti;
	//no machine found with clock init, so use our tcb
	if (min_id==0) {
		min_ti = getTCB(my_id);
		if (min_ti==NULL) {
			printf("Error: Couldn't find me!\n");
			return;
		}
	}
	else {
		min_ti = getTCB(min_id);
		if (min_ti==NULL) {
			printf("Error: Couldn't find TCB %d!\n", min_id);
			return;
		}
	}


	if (!init_state) {
	 if (ti->id == min_ti->id){
		vclock_at_sync = get_vclock_at(min_ti->gclock_info.time_of_sync);
		time_of_sync = min_ti->gclock_info.time_of_sync;
		if (ti->id == my_id)					
			virtual_drift = 0;
		else {	
			virtual_drift = (double)(min_ti->gclock_info.vclock - vclock_at_sync) / SEND_PERIOD;
			apply_drift_until = time_of_sync + SEND_PERIOD;
		}
	 }
	}
	else {
		if (min_ti->gclock_info.vclock!=0){
			if (ti->id == min_ti->id){
				vclock_at_sync = min_ti->gclock_info.vclock;
				time_of_sync = min_ti->gclock_info.time_of_sync;;
				virtual_drift = 0;
			}
		}
	}


}



/************************************************************
 * TCB Services implementation
 ************************************************************/


/**
 * Verifies if a given TCB is subscribed to a given event channel id
 *
 * @param ti TCB info
 * @param ec_id Id of the event channel to verify
 *
 * @return TRUE if subscribed, FALSE otherwise
 */
bool isSubscribed (tcb_info_t* ti, char ec_id) {

	if (ti==NULL) {
		return FALSE;
	}

	//create mask with bit ec_id set
	int mask = 1<<ec_id;
	//check if subscribed
	if (ti->ec_ids&mask)
		return TRUE;
	return FALSE;
}


/**
 * SERVICE: subscribeEventChannel
 *
 * GOAL: to subscribe to a certain event channel.
 *
 * PARAM: ec_id Id of the channel to subscribe. (0<=ec_id<=31)
 *
 * RETURNS (to DLL): 
 *		0 if OK,
 *		-1 if already subscribed,
 *		-2 if ec_id is invalid.
 */
void subscribeEventChannel(char ec_id) {
	int ret=0;

	tcb_info_t* ti = getTCB(my_id);
	if (ti==NULL) {
		printf("Error: Couldn't find me!\n");
		ret=-1;
	}
	//check if ec_id is invalid
	else if (ec_id<0 || ec_id>31)
		ret=-2;
	else {
		EnterCriticalSection(&tcb_critical_section);
		//check if already subscribed
		if (isSubscribed(ti, ec_id))
			ret=-1;
		else {
			 //subscribe
			 int mask = 1<<ec_id;
			 ti->ec_ids|=mask;
			 //debug_tcb_services(("ec_ids= %d\n",ec_ids[my_id]));
			 //printf("ec_ids=%d\n", ec_ids[my_id]);
			 /*
			 printf("ECs subscribed: ");
			 mask = 1;
			 for (int i=0; i<32; i++) {
				if (ti->ec_ids&mask)
					printf("%d ", i);
				mask <<= 1;
			 }
			 printf("\n");
			 */
		}
	}

	LeaveCriticalSection(&tcb_critical_section);

	//fill answer
	srrt->response.service_id=SUBSCRIBE_EVENT_CHANNEL;
	srrt->response.ret=ret;
	srrt->type=2;
	//sinalize dll that answer is ready
	SetEvent(dll_wait_event);
}


/**
 * SERVICE: unsubscribeEventChannel
 *
 * GOAL: to unsubscribe from a certain event channel.
 *
 * PARAM: ec_id Id of the channel to unsubscribe. (0<=ec_id<=31)
 *
 * RETURNS (to DLL): 
 *		0 if OK,
 *		-1 if not subscribed,
 *		-2 if ec_id is invalid.
 */
void unsubscribeEventChannel(char ec_id) {
	int ret=0;

	EnterCriticalSection(&tcb_critical_section);

	tcb_info_t* ti = getTCB(my_id);
	if (ti==NULL) {
		printf("Error: Couldn't find me!\n");
		ret=-1;
	}
	//check if ec_id is invalid
	else if (ec_id<0 || ec_id>31)
		ret=-2;
	else {
		//check if not subscribed
		if (!isSubscribed(ti, ec_id))
			ret=-1;
		else {
			 //unsubscribe
			 int mask = 1<<ec_id;
			 ti->ec_ids&=~mask;
			 //debug_tcb_services(("ec_ids= %d\n",ec_ids[my_id]));
			 //printf("ec_ids=%d\n", ec_ids[my_id]);
			 /*
			 printf("ECs subscribed: ");
			 mask = 1;
			 for (int i=0; i<32; i++) {
				if (ti->ec_ids&mask)
					printf("%d ", i);
				mask <<= 1;
			 }
			 printf("\n");
			 */
		}
	}

	LeaveCriticalSection(&tcb_critical_section);

	//fill answer
	srrt->response.service_id=UNSUBSCRIBE_EVENT_CHANNEL;
	srrt->response.ret=ret;
	srrt->type=2;
	//sinalize dll that answer is ready
	SetEvent(dll_wait_event);
}


/**
 * SERVICE: monEventChannel
 *
 * GOAL: to monitor an event channel QoS.
 *
 * PARAM: 
 *		ec_id Id of the event channel to monitor (0-31)
 *		mode Indicates the monitor mode: 0-stability, 1-awareness
 *		bound Bound (ms) to monitor (>0)
 *		coverage Coverage ([0-1]) to monitor
 *		deviation Maximum deviation of bound/coverage before handler 
 *				  being called
 *		num_samples Number of past samples to consider
 *		samples_source Type of samples to consider: 
 *					   0-all, 1-sent by me, 2-sent by others
 *		dllName Name of the DLL file where the handler is defined
 *		funcName Handler name
 *
 * RETURNS (to DLL):
 *		tag that identifies the monitoring if OK,
 *		-1 if ec_id is invalid,
 *		-2 if mode is invalid,
 *		-3 if bound is not positive,
 *		-4 if coverage is invalid,
 *		-5 if deviation is negative,
 *		-6 if num_samples is not positive OR greater than the maximum number 
 *		   of samples allowed
 *		-7 if samples_source is invalid
 *		-8 if DLL file not found
 *		-9 if handler not found
 *		-10 if there is no available entry in the monitor table
 */
 void monEventChannel(wchar_t ec_id, char mode, int bound, double coverage,
					 double deviation, int num_samples, wchar_t samples_source,
					 wchar_t dllName[50], wchar_t funcName[50]) {

	int ret=0;
	
	if (ec_id<0 || ec_id>31)
		ret=-1;
	else if (mode<0 || mode>1)
		ret=-2;
	else if (bound<=0)
		ret=-3;
	else if (coverage<=0 || coverage>1)
		ret=-4;
	else if (deviation<0)
		ret=-5;
	else if (num_samples<=0 || num_samples>MAX_NUM_SAMPLES)
		ret=-6;
	else if (samples_source<0 || samples_source>2)
		ret=-7;
	else {
		//synchronize with recv thread 
		EnterCriticalSection(&tcb_critical_section);
		
		//find an empty entry on monitor table
		bool found=false;
		for (int i=0; i<MAX_QOS_MONITOR_INFO_TABLE_SIZE && !found; i++) {
			if (qos_monitor_info[i].empty==TRUE) {
				found=true;
			}
		}
		if (!found) {
			ret=-10;
		}
		else {
			i--;
			//load DLL
			HINSTANCE hLib=LoadLibrary(dllName);
			//check if loading was ok
			if(hLib==NULL) {
				ret=-8; 
			}
			else {
				//get function pointer
				qosfunctype funcptr=(qosfunctype)GetProcAddress((HMODULE)hLib, funcName);       
				if((funcptr==NULL)) {
					FreeLibrary((HMODULE)hLib);
					ret=-9;
				}
				//insert new entry in monitor table
				else {
					qos_monitor_info[i].empty = FALSE;
					qos_monitor_info[i].ec_id = (char) ec_id;
					qos_monitor_info[i].mode = mode;
					qos_monitor_info[i].bound = bound;
					qos_monitor_info[i].coverage = coverage;
					qos_monitor_info[i].deviation = deviation;
					qos_monitor_info[i].num_samples = num_samples;
					qos_monitor_info[i].samples_source = samples_source;
					qos_monitor_info[i].handler = funcptr;
					qos_monitor_info[i].pdf.mean = 0;
					qos_monitor_info[i].pdf.variance = 0;
					qos_monitor_info[i].pdf.first_sample = -1;
					qos_monitor_info[i].pdf.last_sample = -1;
					qos_monitor_info[i].pdf.num_samples = 0;
					ret=i;
				}
			} //else (hLib==NULL)
		} //else (!found)
		LeaveCriticalSection(&tcb_critical_section);
	} //else (initial errors)
				
	//fill answer
	srrt->response.service_id=MON_EVENT_CHANNEL;
	srrt->response.ret=ret;
	srrt->type=2;
	//sinalize dll that answer is ready
	SetEvent(dll_wait_event);
 }



/**
 * SERVICE: updateMonEventChannel
 *
 * GOAL: to update a monitoring of an event channel QoS. 
 *		 Updates fields with update=TRUE.
 *
 * PARAM: 
 *		tag Monitoring identifier
 *		ec_id Id of the event channel to monitor (0-31)
 *		mode Indicates the monitor mode: 0-stability, 1-awareness
 *		bound Bound (ms) to monitor (>0)
 *		coverage Coverage ([0-1]) to monitor
 *		deviation Maximum deviation of bound/coverage before handler 
 *				  being called
 *		num_samples Number of past samples to consider
 *		samples_source Type of samples to consider: 
 *					   0-all, 1-sent by me, 2-sent by others
 *
 * RETURNS (to DLL):
 *		0 if OK,
 *		-1 if ec_id is invalid,
 *		-2 if mode is invalid,
 *		-3 if bound is not positive,
 *		-4 if coverage is invalid,
 *		-5 if deviation is negative,
 *		-6 if num_samples is not positive OR greater than the maximum number 
 *		   of samples allowed
 *		-7 if samples_source is invalid
 *		-8 if tag is invalid
 */
 void updateMonEventChannel(int tag, bool update_ec_id, wchar_t ec_id, 
									 bool update_mode, char mode, 
									 bool update_bound, int bound, 
									 bool update_coverage, double coverage,
									 bool update_deviation, double deviation, 
									 bool update_num_samples, int num_samples, 
									 bool update_samples_source, wchar_t samples_source) {

	int ret=0;
	
	if (update_ec_id && (ec_id<0 || ec_id>31))
		ret=-1;
	else if (update_mode && (mode<0 || mode>1))
		ret=-2;
	else if (update_bound && bound<=0)
		ret=-3;
	else if (update_coverage && (coverage<=0 || coverage>1))
		ret=-4;
	else if (update_deviation && deviation<0)
		ret=-5;
	else if (update_num_samples && (num_samples<=0 || num_samples>MAX_NUM_SAMPLES))
		ret=-6;
	else if (update_samples_source && (samples_source<0 || samples_source>2))
		ret=-7;
	else {
		//synchronize with recv thread 
		EnterCriticalSection(&tcb_critical_section);
		
		//verify if tag is valid
		if (tag<0 || tag>=MAX_QOS_MONITOR_INFO_TABLE_SIZE || qos_monitor_info[tag].empty==TRUE) {
			ret=-8;
		}
		else {
			//update entry in monitor table
			if (update_ec_id)
				qos_monitor_info[tag].ec_id = (char) ec_id;
			if (update_mode)
				qos_monitor_info[tag].mode = mode;
			if (update_bound)
				qos_monitor_info[tag].bound = bound;
			if (update_coverage)
				qos_monitor_info[tag].coverage = coverage;
			if (update_deviation)
				qos_monitor_info[tag].deviation = deviation;
			if (update_num_samples)
				qos_monitor_info[tag].num_samples = num_samples;
			if (update_samples_source)
				qos_monitor_info[tag].samples_source = samples_source;
			ret=0;			
		} //else (empty=TRUE)
		LeaveCriticalSection(&tcb_critical_section);
	} //else (initial errors)
				
	//fill answer
	srrt->response.service_id=UPDATE_MON_EVENT_CHANNEL;
	srrt->response.ret=ret;
	srrt->type=2;
	//sinalize dll that answer is ready
	SetEvent(dll_wait_event);
 }


 /**
 * SERVICE: endMonEventChannel
 *
 * GOAL: to end a monitoring of an event channel QoS. 
 *
 * PARAM: 
 *		tag Monitoring identifier
 *
 * RETURNS (to DLL):
 *		0 if OK,
 *		-1 if tag is invalid
 */
 void endMonEventChannel(int tag) {

	int ret=0;
	
	//synchronize with recv thread 
	EnterCriticalSection(&tcb_critical_section);
		
	//verify if tag is valid
	if (tag<0 || tag>=MAX_QOS_MONITOR_INFO_TABLE_SIZE || qos_monitor_info[tag].empty==TRUE) {
		ret=-1;
	}
	else {
		//end monitoring
		qos_monitor_info[tag].empty=TRUE;
	}
		
	LeaveCriticalSection(&tcb_critical_section);
				
	//fill answer
	srrt->response.service_id=END_MON_EVENT_CHANNEL;
	srrt->response.ret=ret;
	srrt->type=2;
	//sinalize dll that answer is ready
	SetEvent(dll_wait_event);
 }




/**
 * SERVICE: getGlobalTimestamp
 *
 * GOAL: to return global clock value in milliseconds
 *		 at the local instant specified. If local instant==0
 *		 the current global clock value is returned.
 *
 * RETURNS (to DLL): 
 *		global clock value in ms at the local instant specified, if localInstant>0,
 *		present global clock value, if localInstant==0.
 */
void getGlobalTimestamp(int localInstant) {
	int timestamp;
	
	//synchronize with recv thread 
	EnterCriticalSection(&tcb_critical_section);
	if (localInstant==0)
		//get current global clock value
		timestamp = get_vclock_at(GetTickCount());
	else
		//get global clock value at local instant
		timestamp = get_vclock_at(localInstant);
	LeaveCriticalSection(&tcb_critical_section);

	debug_tcb_services(("Global Timestamp= %d\n",timestamp));

	//fill answer
	srrt->response.service_id=GET_GLOBAL_TIMESTAMP;
	srrt->response.value=timestamp;
	srrt->type=2;
	//sinalize dll that answer is ready
	SetEvent(dll_wait_event);
}


/**
 * SERVICE: getTimestamp
 *
 * GOAL: to return local clock value in milliseconds
 *
 * RETURNS (to DLL): local clock value in ms
 */
void getTimestamp() {
	int timestamp;
	//get system uptime in milliseconds
	timestamp = GetTickCount();
	debug_tcb_services(("Timestamp= %d\n",timestamp));

	//fill answer
	srrt->response.service_id=GET_TIMESTAMP;
	srrt->response.value=timestamp;
	srrt->type=2;
	//sinalize dll that answer is ready
	SetEvent(dll_wait_event);
}


/**
 * SERVICE: getTCBId
 *
 * GOAL: to return tcb id
 *
 * RETURNS (to DLL): tcb id
 */
void getTCBId() {
	
	debug_tcb_services(("Id= %d\n",my_id));

	//fill answer
	srrt->response.service_id=GET_TCB_ID;
	srrt->response.id = my_id;
	srrt->type=2;
	//sinalize dll that answer is ready
	SetEvent(dll_wait_event);
}



/**
 * SERVICE: startMeasurement
 *
 * GOAL: to start a measurement
 * 
 * PARAM: start_ev Instant that marks the measurement begin
 *
 * RETURNS (to DLL): tag that identifies the measurement, 
 *			-1 if there is no tag available
 */
void startMeasurement(int start_ev){
	int tag=-1;
	debug_tcb_services(("strt_m:start_ev= %d\n",start_ev));
	for (int i=0; (i<MAX_LOCAL_MEASURES && tag==-1); i++) {
		if (localMeasures[i]==0) {
			localMeasures[i]=start_ev;
			tag=i;
		}
	}
	debug_tcb_services(("strt_m:tag= %d\n",tag));
	
	//fill answer
	srrt->response.service_id=START_MEASUREMENT;
	srrt->response.tag=tag;
	srrt->type=2;
	//sinalize dll that answer is ready
	SetEvent(dll_wait_event);
}


/**
 * SERVICE: endMeasurement
 *
 * GOAL: to end a measurement
 * 
 * PARAM: tag Measurement identifier
 *
 * RETURNS (to DLL): Information about the measurement,
 *			end_ev=-1 if tag is invalid
 */
void endMeasurement(int tag){
	measurement_info_t mi;
	debug_tcb_services(("stop_m:tag= %d\n",tag));
	//if tag is invalid, return null
	if (tag<0 || tag>=MAX_LOCAL_MEASURES || localMeasures[tag]==0) {
		mi.end_ev=-1;
	}
	else {
		//get current ime
		int now = GetTickCount();
		mi.end_ev=now;
		//calculate measure
		mi.procdelay=now-localMeasures[tag];
		//free tag
		localMeasures[tag]=0;
		debug_tcb_services(("stop_m:end_ev=%d,procdelay=%d\n",mi.end_ev, mi.procdelay));
	}
	//fill answer
	srrt->response.service_id=END_MEASUREMENT;
	srrt->response.mi=mi;
	srrt->type=2;
	//sinalize dll that answer is ready
	SetEvent(dll_wait_event);
}


/**
 * THREAD: ExecThread
 *
 * GOAL: 1) verify if an application function
 *		 is executable in a specified time interval
 *
 *		 2) if function is executable, execute it
 *
 *		 3) return end instant to DLL
 *
 * RETURNS (to DLL): The time instant when the function ends,
 *					-6 if it's not possible to timely execute function
 *					
 */
DWORD WINAPI ExecThread (LPVOID param) {
	
	int now;
	int ms_since_start_ev;
	int newDelay;
	int specifiedFuncEndInstant;
	int actualFuncEndInstant;
	int end_ev;
	//stores execution table position, passed by timelyExec
	int exec_index = (int)param;

	debug_tcb_services(("timely_th:Thread starting\n"));
	debug_tcb_services(("timely_th:exec_index=%d\n", exec_index));

	//stores current time
	now = GetTickCount();
	//stores number of milliseconds since start_ev
	ms_since_start_ev = now - executions[exec_index].start_ev;
	//recalculates new delay, according to ms_since_start_ev
	if (executions[exec_index].delay < ms_since_start_ev) {
		//if delay ms has already passed
		newDelay=0;
	}
	else {
		//recalculate delay
		newDelay=executions[exec_index].delay-(ms_since_start_ev);
	}
	
	//actual function end instant
	actualFuncEndInstant=now+newDelay+executions[exec_index].wcet;
	//specified function end instant
	specifiedFuncEndInstant=executions[exec_index].start_ev + executions[exec_index].max_exec;

	//if it's not possible to execute function as specified
	if (actualFuncEndInstant > specifiedFuncEndInstant) {
		debug_tcb_services(("timely_th:not timely\n"));
		//fill answer
		srrt->response.service_id=TIMELY_EXEC;
		srrt->response.end_ev=-6;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
		//clean execution table entry
		executions[exec_index].func=NULL;
		//unload DLL
		//FreeLibrary(executions[exec_index].hdll);
		//exit thread
		return 0;
	}

	//if it's possible to timely execute function, 
	//sleep newDelay and then execute function
	if (newDelay>0)
		Sleep(newDelay);
	
	//int now2 = GetTickCount();

	debug_tcb_services(("timely_th:Execute func\n"));

	executions[exec_index].func();
	
	//return end instant
	end_ev=GetTickCount();

	debug_tcb_services1(("timely_th:ms_since_start_ev=%d\n", 
							ms_since_start_ev));
	debug_tcb_services1(("timely_th:Sleeping %d ms\n", newDelay));
	debug_tcb_services1(("timely_th:Actual sleep= %d\n", now2-now));
	debug_tcb_services1(("timely_th:exec time=%d\n", 
						 end_ev-executions[exec_index].start_ev));

	//fill answer
	srrt->response.service_id=TIMELY_EXEC;
	srrt->response.end_ev=end_ev;
	srrt->type=2;
	//sinalize dll that answer is ready
	SetEvent(dll_wait_event);
	//clean execution table entry
	executions[exec_index].func=NULL;
	//unload DLL
	//FreeLibrary(executions[exec_index].hdll);

	//exit thread
	return 0;
}




/**
 * SERVICE: timelyExec
 *
 * GOAL: to start a timely execution
 * 
 * PARAM: 
 *		start_ev A time instant
 *		delay start_ev+delay marks the exec begin
 *		max_exec Deadline for the execution, counting since start_ev
 *		dllName Name of the DLL file where the function is defined
 *		funcName Function name 
 *		wcetName Worst Case Execution Time function name
 *
 * RETURNS (to DLL) : The time instant when the function ends (thread returns)
 *			-1 if no available entry in the executions table
 *			-2 if dll not found
 *			-3 if function not found
 *			-4 if wcet function not found
 *			-5 if it's not possible to execute function
 *			-6 if it's not possible to timely execute function (thread returns)
 */
void timelyExec(int start_ev, int delay, int max_exec, 
			   wchar_t dllName[50], wchar_t funcName[50], wchar_t wcetName[50]){

	//pointer to function
	functype funcptr;
	//pointer to wcet function
	wcettype wcetptr;
	//stores the execution table entry position to be used
	int exec_index=-1;

	
	debug_tcb_services(("timely_e:start_ev=%d, delay=%d, max_exec=%d\n",
						start_ev, delay, max_exec));


	//search for an empty entry in the executions table
	for (int i=0; (i<MAX_EXEC && exec_index==-1); i++) {
		if (executions[i].func==NULL) {
			exec_index = i;
		}
	}
	if (exec_index==-1){
		//fill answer
		srrt->response.service_id=TIMELY_EXEC;
		srrt->response.end_ev=-1;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
		return;
	}

	debug_tcb_services(("timely_e: exec_index=%d\n",exec_index));

	//load DLL
	HINSTANCE hLib=LoadLibrary(dllName);
	//check if loading was ok
    if(hLib==NULL) {
		debug_tcb_services(("timely_e:Dll [%s] not found\n",dllName));
        //fill answer
		srrt->response.service_id=TIMELY_EXEC;
		srrt->response.end_ev=-2;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
		return;
	}

	debug_tcb_services(("timely_e:DLL loaded\n"));

	//get function pointer
    funcptr=(functype)GetProcAddress((HMODULE)hLib, funcName);       
    if((funcptr==NULL)) {
        FreeLibrary((HMODULE)hLib);
		//fill answer
		srrt->response.service_id=TIMELY_EXEC;
		srrt->response.end_ev=-3;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
		return;
    }
	
	debug_tcb_services(("timely_e:Function loaded\n"));

	//get function wcet pointer
    wcetptr=(wcettype)GetProcAddress((HMODULE)hLib, wcetName);       
    if((wcetptr==NULL)) {
        FreeLibrary((HMODULE)hLib);
		//fill answer
		srrt->response.service_id=TIMELY_EXEC;
		srrt->response.end_ev=-4;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
        return;
    }

	debug_tcb_services(("timely_e:WCET Function loaded\n"));

	//fill executions table entry
	executions[exec_index].func = funcptr;
	executions[exec_index].wcet = wcetptr();
	executions[exec_index].start_ev = start_ev;
	executions[exec_index].delay = delay;
	executions[exec_index].max_exec = max_exec;
	executions[exec_index].hdll = (HMODULE)hLib;
	//create thread that will execute function
	HANDLE hThread =
					CreateThread(NULL,0,ExecThread,(void*)exec_index,CREATE_SUSPENDED,NULL);
	if (hThread==NULL){
		FreeLibrary((HMODULE)hLib);
		//fill answer
		srrt->response.service_id=TIMELY_EXEC;
		srrt->response.end_ev=-5;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
		return;
	}
	//set thread priority to MAXIMUM-2
	CeSetThreadPriority(hThread, 2);
	//start thread
	ResumeThread(hThread);

	debug_tcb_services(("timely_e:Thread created and started\n"));
}




/**
 * THREAD: LocalFDThread
 *
 * GOAL: to verify if a timing failure occurs 
 *		 and if it occurs, call the handler recved
 *
 * RETURNS (to DLL) : tag that identifies the timing failure detection, 
 *					  -6 if a timing failure already occured.
 */
DWORD WINAPI LocalFDThread (LPVOID param) {
	
	int now;
	int ms_since_start_ev;
	//stores localFD table position, passed by startLocalFD
	int localFD_index = (int)param;

	debug_tcb_services(("localFD_th:Thread starting\n"));
	debug_tcb_services(("localFD_th:localFD_index=%d\n", localFD_index));

	//stores current time
	now = GetTickCount();
	//stores number of milliseconds since start_ev
	ms_since_start_ev = now - localFD[localFD_index].start_ev;
	
	//verify if failure already occured
	if (localFD[localFD_index].spec <= ms_since_start_ev) {
		//if yes, return error to DLL and exit thread
		debug_tcb_services(("localFD_th:already failed!\n"));
		//fill answer
		srrt->response.service_id=START_LOCAL_FD;
		srrt->response.tag=-6;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
		//clean localFD table entry
		localFD[localFD_index].func=NULL;
		//unload DLL
		FreeLibrary(localFD[localFD_index].hdll);
		//exit thread
		return 0;
	}
	
	//otherwise, return tag to DLL
	debug_tcb_services(("localFD_th:return tag=%d\n",localFD_index));
	//fill answer
	srrt->response.service_id=START_LOCAL_FD;
	srrt->response.tag=localFD_index;
	srrt->type=2;
	//sinalize dll that answer is ready
	SetEvent(dll_wait_event);
	
	//sleep for (spec - ms_since_start_ev) milliseconds
	debug_tcb_services(("localFD_th:Sleeping %d ms\n", 
			localFD[localFD_index].spec - ms_since_start_ev));
	Sleep(localFD[localFD_index].spec - ms_since_start_ev);

	/* during sleep, this thread can be stopped by endLocalFD */

	//if it reaches here, it means that a failure has occured
	debug_tcb_services(("localFD_th:Failure!, exec handler\n"));
	//this allows endLocalFD to know that a failure has occured
	localFD[localFD_index].faulty=TRUE;
	//execute failure handler
	localFD[localFD_index].func();
	
	//exit thread
	return 0;
}



/**
 * SERVICE: startLocalFD
 *
 * GOAL: to start a local timing failure detection
 * 
 * PARAM: 
 *		start_ev A time instant
 *		spec A timing failure occurs if start_ev+spec instant is reached
 *			 before a call endLocalFD
 *		deadline Deadline for the execution of handler, if a failure occurs
 *		dllName Name of the DLL file where the handler is defined
 *		funcName Handler name 
 *		wcetName Handler's Worst Case Execution Time function name
 *
 * RETURNS (to DLL) : tag that identifies the timing failure detection 
 *					  (thread returns),
 *			-1 if no available entry in the localFD table,
 *			-2 if dll not found,
 *			-3 if function not found,
 *			-4 if wcet function not found,
 *			-5 if it's not possible to execute timing failure detection,
 *			-6 if a timing failure already occured (thread returns).
 */
void startLocalFD(int start_ev, int spec, int deadline, 
			   wchar_t dllName[50], wchar_t funcName[50], wchar_t wcetName[50]){

	//pointer to function
	functype funcptr;
	//pointer to wcet function
	wcettype wcetptr;
	//stores the localFD table entry position to be used
	int localFD_index=-1;

	
	debug_tcb_services(("s_localFD:start_ev=%d, spec=%d, deadline=%d\n",
						start_ev, spec, deadline));


	//search for an empty entry in the executions table
	for (int i=0; (i<MAX_LOCAL_FD && localFD_index==-1); i++) {
		if (localFD[i].func==NULL) {
			localFD_index = i;
		}
	}
	if (localFD_index==-1){
		//fill answer
		srrt->response.service_id=START_LOCAL_FD;
		srrt->response.tag=-1;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
		return;
	}

	debug_tcb_services(("s_localFD: localFD_index=%d\n",localFD_index));

	//load DLL
	HINSTANCE hLib=LoadLibrary(dllName);
	//check if loading was ok
    if(hLib==NULL) {
		debug_tcb_services(("s_localFD:Dll [%s] not found\n",dllName));
        //fill answer
		srrt->response.service_id=START_LOCAL_FD;
		srrt->response.tag=-2;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
		return;
	}

	debug_tcb_services(("s_localFD:DLL loaded\n"));

	//get function pointer
    funcptr=(functype)GetProcAddress((HMODULE)hLib, funcName);       
    if((funcptr==NULL)) {
        FreeLibrary((HMODULE)hLib);
		//fill answer
		srrt->response.service_id=START_LOCAL_FD;
		srrt->response.tag=-3;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
		return;
    }
	
	debug_tcb_services(("s_localFD:Function loaded\n"));

	//get function wcet pointer (although, currently it's not used)
    wcetptr=(wcettype)GetProcAddress((HMODULE)hLib, wcetName);       
    if((wcetptr==NULL)) {
        FreeLibrary((HMODULE)hLib);
		//fill answer
		srrt->response.service_id=START_LOCAL_FD;
		srrt->response.tag=-4;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
        return;
    }

	debug_tcb_services(("s_localFD:WCET Function loaded\n"));

	//fill localFD table entry 
	localFD[localFD_index].func = funcptr;
	localFD[localFD_index].faulty = FALSE;
	localFD[localFD_index].start_ev = start_ev;
	localFD[localFD_index].spec = spec;
	localFD[localFD_index].deadline = deadline;
	localFD[localFD_index].hdll = (HMODULE)hLib;
	//create and start (normal priority) thread that will monitor timing failure
	localFD[localFD_index].task_id =
				CreateThread(NULL,0,LocalFDThread,(void*)localFD_index,CREATE_SUSPENDED,NULL);
	if (localFD[localFD_index].task_id==NULL){
		FreeLibrary((HMODULE)hLib);
		//fill answer
		srrt->response.service_id=START_LOCAL_FD;
		srrt->response.tag=-5;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
	}
	//set thread priority to MAXIMUM-2
	CeSetThreadPriority(localFD[localFD_index].task_id, 2);
	//start thread
	ResumeThread(localFD[localFD_index].task_id);

	debug_tcb_services(("s_localFD:Thread created and started\n"));
}


/**
 * SERVICE: endLocalFD
 *
 * GOAL: to end a local timing failure detection
 * 
 * PARAM: tag Local timing detection identifier
 *
 * RETURNS (to DLL): Information about the timing failure detection,
 *					 end_ev=-1 if tag is invalid
 */
void endLocalFD (int tag){
	debug_tcb_services(("stop_lfd:tag= %d\n",tag));
	//if tag is invalid, return null
	if (tag<0 || tag>=MAX_LOCAL_FD || localFD[tag].func==0) {
		srrt->response.endLocalFDRet.end_ev=-1;
	}
	else {
		//stop timing failure detection if it's still running
		if (!localFD[tag].faulty) {
			TerminateThread(localFD[tag].task_id,0);
		}
		//close thread handle to free memory
		CloseHandle(localFD[tag].task_id);

		int now = GetTickCount();
		srrt->response.endLocalFDRet.end_ev = now;
		//number of milliseconds since start_ev
		srrt->response.endLocalFDRet.procdelay = now-localFD[tag].start_ev;
		srrt->response.endLocalFDRet.faulty = localFD[tag].faulty;
	}
	//fill answer
	srrt->response.service_id=END_LOCAL_FD;
	srrt->type=2;
	//sinalize TCB dll that answer is ready
	SetEvent(dll_wait_event);
	//clean localFD table entry
	localFD[tag].func=NULL;
	//unload handler DLL
	//FreeLibrary(localFD[tag].hdll);
}



/**
 * THREAD: DistFDThread
 *
 * GOAL: to call a distributed failure handler, after
 *		 sleeping the release_time specified
 *
 */
DWORD WINAPI DistFDThread (LPVOID param) {
	
	//arguments are in a global variable: dfi

	debug_tcb_services(("distFD_th:Thread starting\n"));
	debug_tcb_services(("distFD_th:failure_at_time=%d\n", dfi.failure_at_time));

	//sleeps release_time specified
	Sleep(dfi.release_time);

	//executes failure handler
	dfi.funcptr(dfi.from, dfi.seq);
	
	//exit thread
	return 0;
}



/**
 * SERVICE: startDistributedFD
 *
 * GOAL: to start a distributed timing failure detection
 * 
 * PARAM: 
 *		ec_id Event channel where the timing failure detection will be
 *				  started (0<=event_channel_id<=31).
 *		send_ev The initial time instant of the failure detection.
 *		spec A timing failure occurs if start_ev+spec instant is reached
 *			 before a call to receive in destination occurs
 *		deadline Deadline for the execution of handler, if a failure occurs
 *		dllName Name of the DLL file where the handler is defined
 *		funcName Handler name 
 *		wcetName Handler's Worst Case Execution Time function name
 *
 * RETURNS (to DLL) : seq that identifies the dist. timing failure detection,
 *			-2 if dll not found,
 *			-3 if function not found,
 *			-4 if wcet function not found,
 *			-5 if there are too many distributed fds already running,
 *			-6 if a timing failure already occured,
 *			-7 if it's not possible to execute timing failure detection,
 *			-8 if no TCB is subscribed to event channel specified.   
 */
void startDistributedFD(wchar_t ec_id, int send_ev, int spec, 
				  int deadline, wchar_t dllName[50], wchar_t funcName[50], 
				  wchar_t wcetName[50]){

	//pointer to DLL
	HINSTANCE hLib=NULL;
	//pointer to function
	distfunctype funcptr;
	//pointer to wcet function
	wcettype wcetptr;
	//indicates the recipients of msg (indexed to membership)
	bool to_array[MAX_NUMBER_OF_MACHINES];
	//total number of recipients
	int number_of_dest=0;
	//indicates the send info table entry used
	int sndPos=0;
	
	
	debug_tcb_services(("s_distFD:send_ev=%d, spec=%d, deadline=%d\n",
						send_ev, spec, deadline));

	/* verify if MAX_DISTRIBUTED_FD_SERVICES has been reached 
	   in this slot */
	if (n_sends_in_this_slot == MAX_DISTRIBUTED_FD_SERVICES) {
		//fill answer
		srrt->response.service_id=START_DISTRIBUTED_FD;
		srrt->response.seq=-5;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
		return;
	}

	//next data is accessed by other threads
	EnterCriticalSection(&tcb_critical_section);

	//init to_array
	for (int i=0; i<MAX_NUMBER_OF_MACHINES; i++)
		to_array[i] = FALSE;

	/* analise ec_id and conclude who are the recipients */
	for (i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		if (membership[i].empty)
			continue;

		if (isSubscribed(&membership[i], ec_id)) {
			to_array[i]=TRUE;
			number_of_dest++;
		}
	}
	
	//if no TCB is subscribed to event channel specified
	if (number_of_dest==0) {
		//fill answer
		srrt->response.service_id=START_DISTRIBUTED_FD;
		srrt->response.seq=-8;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
		LeaveCriticalSection(&tcb_critical_section);
		return;
	}

	/* verify if dll is specified. if not, it means that no failure handler
	   should be executed if a timing failure occurs */
	if (wcscmp(dllName,L"")!=0) {
		//load DLL
		hLib=LoadLibrary(dllName);
		//check if loading was ok
		if(hLib==NULL) {
			debug_tcb_services(("s_distFD:Dll [%s] not found\n",dllName));
			//fill answer
			srrt->response.service_id=START_DISTRIBUTED_FD;
			srrt->response.seq=-2;
			srrt->type=2;
			//sinalize dll that answer is ready
			SetEvent(dll_wait_event);
			LeaveCriticalSection(&tcb_critical_section);
			return;
		}

		debug_tcb_services(("s_distFD:DLL loaded\n"));

		//get function pointer
		funcptr=(distfunctype)GetProcAddress((HMODULE)hLib, funcName);       
		if((funcptr==NULL)) {
			FreeLibrary((HMODULE)hLib);
			//fill answer
			srrt->response.service_id=START_DISTRIBUTED_FD;
			srrt->response.seq=-3;
			srrt->type=2;
			//sinalize dll that answer is ready
			SetEvent(dll_wait_event);
			LeaveCriticalSection(&tcb_critical_section);
			return;
		}
	
		debug_tcb_services(("s_distFD:Function loaded\n"));

		//get function wcet pointer (although, currently it's not used)
		wcetptr=(wcettype)GetProcAddress((HMODULE)hLib, wcetName);       
		if((wcetptr==NULL)) {
		    FreeLibrary((HMODULE)hLib);
			//fill answer
			srrt->response.service_id=START_DISTRIBUTED_FD;
			srrt->response.seq=-4;
			srrt->type=2;
			//sinalize dll that answer is ready
			SetEvent(dll_wait_event);
			LeaveCriticalSection(&tcb_critical_section);
			return;
		}

		debug_tcb_services(("s_distFD:WCET Function loaded\n"));
		
	} 
	else funcptr = NULL;

	

	bool error=FALSE;
	/* verify if MAX_DISTRIBUTED_FD_SERVICES has been reached */
	for (i=0; i<MAX_NUMBER_OF_MACHINES && !error; i++) {
		if (to_array[i]) {
			if (membership[i].number_of_detections+1>MAX_DISTRIBUTED_FD_SERVICES) {
					error=TRUE;		
			}
		}
	}
	if (error) {
		//return error to DLL
		LeaveCriticalSection(&tcb_critical_section);
		if (funcptr!=NULL)
			FreeLibrary((HMODULE)hLib);
		//fill answer
		srrt->response.service_id=START_DISTRIBUTED_FD;
		srrt->response.seq=-5;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
		LeaveCriticalSection(&tcb_critical_section);
		return;
	}

	
	//update number of detections for each recipient
	for (i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		if (to_array[i]){
			membership[i].number_of_detections++;
		}
	}
	
	//update number of detections in the current send period
	n_sends_in_this_slot++;

	bool found_entry=FALSE;

	//find a free entry on the send_info table
	for (i=0; i<MAX_SEND_INFO_TABLE_SIZE && !found_entry; i++) {
		if (send_info[i].time_of_generation == -1) {
			found_entry=TRUE;
			send_info[i].time_of_generation = send_ev;
			send_info[i].spec = spec;
			send_info[i].ec_id = ec_id;
			//fill destinations
			for (int j=0; j<MAX_NUMBER_OF_MACHINES; j++) {
				if (to_array[j])
					send_info[i].destination[j] = membership[j].id;
				else
					send_info[i].destination[j] = 0;
			}
			send_info[i].remote_fd = FALSE;
			//tcb_seq_num is increased below
			send_info[i].seq = tcb_seq_num;
			//store entry index for possible delete below
			sndPos = i;

#ifdef LOG_QUEUES
		/*
			//if there is a <from,seq-1>, reuse that entry
			bool found=FALSE;
			for (int k=0; k<send_log_size && !found; k++) {
				if (send_log[k].from == my_id &&
					send_log[k].seq == tcb_seq_num-1) {
				
					found = TRUE;
					send_log[k].seq = tcb_seq_num;
				}

			}
			
			if (!found) {
				send_log[send_log_size].from = my_id;
				send_log[send_log_size++].seq = tcb_seq_num;
			}
		*/
#endif

		}
	}

	if (!found_entry) {
		//if the system is correct, never enters here
		printf("FATAL ERROR: No free send info entry\n");
		wait_user();
	}
	
	found_entry = FALSE;
	
	//find a free entry on the wait_info table
	for (i=0; i<MAX_WAIT_INFO_TABLE_SIZE && !found_entry; i++) {
		if (wait_info[i].failure_at_time==-1) {
			found_entry = TRUE;
			//if a failure handler was specified, init failTask
			if (funcptr!=NULL) {
				/* prepare arguments for failTask */
				//dfi is a global variable
				dfi.funcptr = funcptr;
				dfi.deadline = deadline;
				dfi.from = my_id;
				dfi.seq = tcb_seq_num;
				//for now, this value is not used
				dfi.release_time=0;
				dfi.failure_at_time = send_ev+spec;
				//verify if a failure has already occured
				if (dfi.failure_at_time<GetTickCount()) {
					//delete send info entry
					send_info[sndPos].time_of_generation=-1;
					//decrement counters
					n_sends_in_this_slot--;
					for (i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
						if (to_array[i]){
							membership[i].number_of_detections--;
						}
					}
					LeaveCriticalSection(&tcb_critical_section);
					//return error to DLL
					if (funcptr!=NULL)
						FreeLibrary((HMODULE)hLib);
					//fill answer
					srrt->response.service_id=START_DISTRIBUTED_FD;
					srrt->response.seq=-6;
					srrt->type=2;
					//sinalize dll that answer is ready
					SetEvent(dll_wait_event);
					return;
				}
				//create thread that will monitor timing failure
				wait_info[i].task_id =
					CreateThread(NULL,0,DistFDThread,NULL,CREATE_SUSPENDED,NULL);
				if (wait_info[i].task_id==NULL){
					//delete send info entry
					send_info[sndPos].time_of_generation=-1;
					//decrement counters
					n_sends_in_this_slot--;
					for (i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
						if (to_array[i]) {
							membership[i].number_of_detections--;
						}
					}
					LeaveCriticalSection(&tcb_critical_section);
					//return error to DLL
					if (funcptr!=NULL)
						FreeLibrary((HMODULE)hLib);
					//fill answer
					srrt->response.service_id=START_DISTRIBUTED_FD;
					srrt->response.seq=-7;
					srrt->type=2;
					//sinalize dll that answer is ready
					SetEvent(dll_wait_event);
					return;
				}
				//set thread priority to MAXIMUM-2
				CeSetThreadPriority(wait_info[i].task_id, 2);
				//thread is only started if a failure occurs

			} //endif(funcptr!=NULL)
			else {
				wait_info[i].task_id = NULL;
			}

			//fill wait_info entry
			wait_info[i].failure_at_time = send_ev+spec;
			wait_info[i].seq = tcb_seq_num;
			//fill destinations
			for (int j=0; j<MAX_NUMBER_OF_MACHINES; j++) {
				if (to_array[j])
					wait_info[i].destination[j] = membership[j].id;
				else
					wait_info[i].destination[j] = 0;
			}
			wait_info[i].n_dests = (char)number_of_dest;
			wait_info[i].any_fail = FALSE;
			wait_info[i].from = my_id;
			wait_info[i].remote_fd = FALSE;
			wait_info[i].ec_id = ec_id;
			wait_info[i].spec = spec;
			for (j=0; j<MAX_NUMBER_OF_MACHINES; j++) {
				if (to_array[j]) {
					wait_info[i].failure_info[j].empty=FALSE;
					wait_info[i].failure_info[j].id = membership[j].id;
					//2=no info
					wait_info[i].failure_info[j].fail=2;
				}
				else
					wait_info[i].failure_info[j].empty=TRUE;
			}

/*
#ifdef LOG_QUEUES
			
			//if there is a <from,seq-1>, reuse that entry
			bool found=FALSE;
			for (int k=0; k<wait_log_size && !found; k++) {
				if (wait_log[k].from == my_id &&
					wait_log[k].seq == tcb_seq_num-1) {
						found = TRUE;
						wait_log[k].seq = tcb_seq_num;
				}
			}	
			
			if (!found) {
				wait_log[wait_log_size].from = my_id;
				wait_log[wait_log_size++].seq = tcb_seq_num;
			}
			
#endif
*/

		} //endif(wait_info[i].failure_at_time==-1)
	} //endfor

	if (!found_entry) {
		//if the system is correct, never enters here
		printf("FATAL ERROR: No free wait info entry\n");
		wait_user();
	}
	LeaveCriticalSection(&tcb_critical_section);
	
	/*return seq number to application*/
	//fill answer
	srrt->response.service_id=START_DISTRIBUTED_FD;
	srrt->response.seq=tcb_seq_num;
	srrt->type=2;
	//sinalize dll that answer is ready
	SetEvent(dll_wait_event);

	tcb_seq_num++;
	
	//if counter reaches limit, don't let it has value 0 because
	//seq=0 means msg with no distributed FD
	if (tcb_seq_num==2147483647)
		tcb_seq_num=1;
	
}



/**
 * SERVICE: startDistributedRemoteFD
 *
 * GOAL: to start a distributed remote timing failure detection
 * 
 * PARAM: 
 *		ec_id Event channel where the timing failure detection will be
 *				  started (0<=event_channel_id<=31).
 *		send_ev The initial time instant of the failure detection.
 *		spec A timing failure occurs if start_ev+spec instant is reached
 *			 before a call to receive in destination occurs
 *
 * RETURNS (to DLL) : seq that identifies the dist. timing failure detection,
 *			-5 if there are too many distributed fds already running,
 *			-6 if a timing failure already occured,
 *			-7 if it's not possible to execute timing failure detection,
 *			   
 */
void startDistributedRemoteFD(wchar_t ec_id, int send_ev, int spec){

	
	//indicates the recipients of msg
	bool to_array[MAX_NUMBER_OF_MACHINES];
	//total number of recipients
	int number_of_dest=0;
	//indicates the send info table entry used
	int sndPos=0;
	
	
	debug_tcb_services(("s_distFD:send_ev=%d, spec=%d, deadline=%d\n",
						send_ev, spec, deadline));

	/* verify if MAX_DISTRIBUTED_FD_SERVICES has been reached 
	   in this slot */
	if (n_sends_in_this_slot == MAX_DISTRIBUTED_FD_SERVICES) {
		printf("n_sends_in_this_slot=MAX_DISTRIBUTED_FD_SERVICES\n");
		//fill answer
		srrt->response.service_id=START_DISTRIBUTED_FD;
		srrt->response.seq=-5;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
		return;
	}

	//next data is accessed by other threads
	EnterCriticalSection(&tcb_critical_section);

	//init to_array
	for (int i=0; i<MAX_NUMBER_OF_MACHINES; i++)
		to_array[i] = FALSE;


	/* analise ec_id and conclude who are the recipients */
	for (i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		if (membership[i].empty)
			continue;

		if (isSubscribed(&membership[i], ec_id)) {
			to_array[i]=TRUE;
			number_of_dest++;
		}
	}
	
	//if no TCB is subscribed to event channel specified
	if (number_of_dest==0) {
		//fill answer
		srrt->response.service_id=START_DISTRIBUTED_FD;
		srrt->response.seq=-8;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
		LeaveCriticalSection(&tcb_critical_section);
		return;
	}


	bool error=FALSE;
	/* verify if MAX_DISTRIBUTED_FD_SERVICES has been reached */
	for (i=0; i<MAX_NUMBER_OF_MACHINES && !error; i++) {
		if (to_array[i]) {
			if (membership[i].number_of_detections+1>MAX_DISTRIBUTED_FD_SERVICES) {
					printf("number_of_detections[%d]=MAX_DISTRIBUTED_FD_SERVICES\n",i);
					error=TRUE;		
			}
		}
	}
	if (error) {
		//return error to DLL
		LeaveCriticalSection(&tcb_critical_section);
		//fill answer
		srrt->response.service_id=START_DISTRIBUTED_FD;
		srrt->response.seq=-5;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
		return;
	}

	
	//update number of detections for each recipient
	for (i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		if (to_array[i]){
			membership[i].number_of_detections++;
		}
	}
	//update number of detections in the current send period
	n_sends_in_this_slot++;

	bool found_entry=FALSE;

	//find a free entry on the send_info table
	for (i=0; i<MAX_SEND_INFO_TABLE_SIZE && !found_entry; i++) {
		if (send_info[i].time_of_generation == -1) {
			found_entry=TRUE;
			send_info[i].time_of_generation = send_ev;
			send_info[i].spec = spec;
			send_info[i].ec_id = ec_id;

			//fill destinations
			for (int j=0; j<MAX_NUMBER_OF_MACHINES; j++) {
				if (to_array[j])
					send_info[i].destination[j] = membership[j].id;
				else
					send_info[i].destination[j] = 0;
			}
			send_info[i].remote_fd = TRUE;
			//tcb_seq_num is increased below
			send_info[i].seq = tcb_seq_num;
			//store entry index for possible delete below
			sndPos = i;

#ifdef LOG_QUEUES
			/*
			//if there is a <from,seq-1>, reuse that entry
			bool found=FALSE;
			for (int k=0; k<send_log_size && !found; k++) {
				if (send_log[k].from == my_id &&
					send_log[k].seq == tcb_seq_num-1) {
				
					found = TRUE;
					send_log[k].seq = tcb_seq_num;
				}

			}
			
			if (!found) {
				send_log[send_log_size].from = my_id;
				send_log[send_log_size++].seq = tcb_seq_num;
			}
		*/
#endif
		}
	}

	if (!found_entry) {
		//if the system is correct, never enters here
		printf("FATAL ERROR: No free send info entry\n");
		wait_user();
	}
	
	found_entry = FALSE;
	
	//find a free entry on the wait_info table
	for (i=0; i<MAX_WAIT_INFO_TABLE_SIZE && !found_entry; i++) {
		if (wait_info[i].failure_at_time==-1) {
			found_entry = TRUE;
			//it's a remote FD, so don't specify a handler
			wait_info[i].task_id = NULL;
			//fill wait_info entry
			wait_info[i].failure_at_time = send_ev+spec;
			wait_info[i].seq = tcb_seq_num;
			//fill destinations
			for (int j=0; j<MAX_NUMBER_OF_MACHINES; j++) {
				if (to_array[j])
					wait_info[i].destination[j] = membership[j].id;
				else
					wait_info[i].destination[j] = 0;
			}
			wait_info[i].n_dests = number_of_dest;
			wait_info[i].any_fail = FALSE;
			wait_info[i].from = my_id;
			wait_info[i].remote_fd = TRUE;
			wait_info[i].spec = spec;
			wait_info[i].ec_id = ec_id;
			for (j=0; j<MAX_NUMBER_OF_MACHINES; j++) {
				if (to_array[j]) {
					wait_info[i].failure_info[j].empty=FALSE;
					wait_info[i].failure_info[j].id = membership[j].id;
					//2=no info
					wait_info[i].failure_info[j].fail=2;
				}
				else
					wait_info[i].failure_info[j].empty=TRUE;
			}
/*
#ifdef LOG_QUEUES

			//if there is a <from,seq-1>, reuse that entry
			bool found=FALSE;
			for (int k=0; k<wait_log_size && !found; k++) {
				if (wait_log[k].from == my_id &&
					wait_log[k].seq == tcb_seq_num-1) {
						found = TRUE;
						wait_log[k].seq = tcb_seq_num;
				}
			}	
			
			if (!found) {
				wait_log[wait_log_size].from = my_id;
				wait_log[wait_log_size++].seq = tcb_seq_num;
			}
#endif
*/
		} //endif(wait_info[i].failure_at_time==-1)
	} //endfor

	if (!found_entry) {
		//if the system is correct, never enters here
		printf("FATAL ERROR: No free wait info entry\n");
		wait_user();
	}
	LeaveCriticalSection(&tcb_critical_section);
	
	/*return seq number to application*/
	//fill answer
	srrt->response.service_id=START_DISTRIBUTED_REMOTE_FD;
	srrt->response.seq=tcb_seq_num;
	srrt->type=2;
	//sinalize dll that answer is ready
	SetEvent(dll_wait_event);

	tcb_seq_num++;
	
	//if counter reaches limit, don't let it has value 0 because
	//seq=0 means msg with no distributed FD
	if (tcb_seq_num==2147483647)
		tcb_seq_num=1;
	
}




/**
 * SERVICE: defineRemoteFDHandler
 *
 * GOAL: to define the handler to be used on remote failure detection
 * 
 * PARAM: 
 *		dllName Name of the DLL file where the handler is defined
 *		funcName Handler name 
 *		wcetName Handler's Worst Case Execution Time function name
 *		deadline Maximum time allowed for the execution of the handler
 *				 (if a failure occurs)				
 *
 * RETURNS (to DLL) : 0 if OK,
 *			-2 if dll not found,
 *			-3 if function not found,
 *			-4 if wcet function not found,
 *			   
 */
void defineRemoteFDHandler(wchar_t dllName[50], wchar_t funcName[50], 
						   wchar_t wcetName[50], int deadline){


	/* currently, deadline is not used */

	//pointer to DLL
	HINSTANCE hLib=NULL;
	//pointer to function
	distfunctype funcptr;
	//pointer to wcet function
	wcettype wcetptr;
	
	/* verify if dll is specified. if not, it means that no failure handler
	   should be executed if a timing failure occurs */
	if (wcscmp(dllName,L"")!=0) {
		//load DLL
		hLib=LoadLibrary(dllName);
		//check if loading was ok
		if(hLib==NULL) {
			//fill answer
			srrt->response.service_id=DEFINE_REMOTE_HANDLER;
			srrt->response.seq=-2;
			srrt->type=2;
			//sinalize dll that answer is ready
			SetEvent(dll_wait_event);
			return;
		}

		//get function pointer
		funcptr=(distfunctype)GetProcAddress((HMODULE)hLib, funcName);       
		if((funcptr==NULL)) {
			FreeLibrary((HMODULE)hLib);
			//fill answer
			srrt->response.service_id=DEFINE_REMOTE_HANDLER;
			srrt->response.seq=-3;
			srrt->type=2;
			//sinalize dll that answer is ready
			SetEvent(dll_wait_event);
			return;
		}
	
		//get function wcet pointer (although, currently it's not used)
		wcetptr=(wcettype)GetProcAddress((HMODULE)hLib, wcetName);       
		if((wcetptr==NULL)) {
		    FreeLibrary((HMODULE)hLib);
			//fill answer
			srrt->response.service_id=DEFINE_REMOTE_HANDLER;
			srrt->response.seq=-4;
			srrt->type=2;
			//sinalize dll that answer is ready
			SetEvent(dll_wait_event);
			return;
		}		
	} 
	else funcptr = NULL;

	//store the handler for future use
	remote_fd_handler = funcptr;
	remote_fd_deadline = deadline;

	/*return OK to application*/
	//fill answer
	srrt->response.service_id=DEFINE_REMOTE_HANDLER;
	srrt->response.ret=0;
	srrt->type=2;
	//sinalize dll that answer is ready
	SetEvent(dll_wait_event);
}


/**
 * FUNCTION: is_already_in_rcv_info
 *
 * GOAL: to verify if a msg received on payload was already received
 *		 in the control channel. If yes, checks if it was timely or not.
 *
 * PARAM: 
 *		seq - sequence number
 *		from - sender id
 *		time_of_receipt - reception time instant
 *
 * RETURNS: TRUE if it was already received in the control channel,
 *		    FALSE otherwise
 */
bool is_already_in_rcv_info (int seq, unsigned long from, int time_of_receipt) {

	debug_tcb(("is_already_rcv_info(from=%d,seq=%d)\n",from, seq));

	bool found_entry = FALSE;

	//search for msg with (from,seq) received in receive_info
	for (int i=0; i<MAX_RCV_INFO_TABLE_SIZE && !found_entry; i++) {
		if (receive_info[i].failure_at_time!=-1 &&
			receive_info[i].from == from &&
			receive_info[i].seq == seq) {
			//check if it was timely or not
			if (time_of_receipt>receive_info[i].failure_at_time) {
				//timing failure!
				receive_info[i].fail=1;
				receive_info[i].lateness_avail=true;
				receive_info[i].lateness_degree=
					time_of_receipt-receive_info[i].failure_at_time;
			}
			else {
				//timely
				receive_info[i].fail=0;
				receive_info[i].lateness_avail=true;
				receive_info[i].lateness_degree=
					receive_info[i].failure_at_time-time_of_receipt;
			}
			found_entry=TRUE;
		}
	}

	return found_entry;
}


/**
 * PSEUDO-SERVICE: msg_rcv_on_payload
 *
 * GOAL: to DLL notify TCB that a msg was received on payload
 * 
 * PARAM: 
 *		seq - sequence number
 *		from - sender id
 *		time_of_receipt - reception time instant. If time_of_receipt=-2,
 *						  only pre-allocate a position in receive_payload_info.
 *
 * RETURNS (to DLL) : TRUE if msg can be delivered,
 *					  FALSE if msg cannot be delivered			   
 */
void msgRcvOnPayload(int seq, unsigned long from, int time_of_receipt){
	
	debug_tcb(("msg_rcv_pload(from=%d,seq=%d)\n",from, seq));

	EnterCriticalSection(&tcb_critical_section);

	int current_time = GetTickCount();

	//pre-allocate a position in receive_payload_info
	if (time_of_receipt==-2) {
		bool found_entry = FALSE;
		for (int i=0; i<MAX_RCV_PLOAD_TABLE_SIZE && !found_entry; i++) {
			if (receive_payload_info[i].time_of_receipt==-1 || 
				current_time > receive_payload_info[i].clean_up_after) {

				receive_payload_info[i].from = from;
				receive_payload_info[i].seq = seq;
				receive_payload_info[i].time_of_receipt = time_of_receipt;
				receive_payload_info[i].clean_up_after =
							current_time+2*(SEND_PERIOD+RECV_PERIOD);
				found_entry=TRUE;
			}
		}

		LeaveCriticalSection(&tcb_critical_section);

		//if an entry was filled, return TRUE; otherwise return FALSE
		if (found_entry) {
			srrt->response.service_id=MSG_RCV_ON_PAYLOAD;
			srrt->response.deliver=TRUE;
			srrt->type=2;
			//sinalize dll that answer is ready
			SetEvent(dll_wait_event);
		}
		else {
			//in this situation, msg is not delivered to application
			//to guarantee system correction
			srrt->response.service_id=MSG_RCV_ON_PAYLOAD;
			srrt->response.deliver=FALSE;
			srrt->type=2;
			//sinalize dll that answer is ready
			SetEvent(dll_wait_event);
		}
		return;
	}


	//else, do normal operations

	//verify if already received FD info in the control channel
	if (!is_already_in_rcv_info(seq, from, time_of_receipt)){
		
		//if FD info was not yet received in the control channel,
		//fill an entry of the receive_payload_info table. An entry
		//can be filled if it's empty (time_of_receipt==-1) or if 
		//it's "out of date" or if it was pre-allocated
		bool found_entry = FALSE;
		
		//verify if there is a pre-allocated receive_payload_info 
		//table entry and fill it
		for (int i=0; i<MAX_RCV_PLOAD_TABLE_SIZE && !found_entry; i++) {
			if (
				/* verify if it's pre-allocated */
				(receive_payload_info[i].time_of_receipt == -2 &&
				 receive_payload_info[i].from == from &&
				 receive_payload_info[i].seq == seq) ) {

				receive_payload_info[i].from = from;
				receive_payload_info[i].seq = seq;
				receive_payload_info[i].time_of_receipt = time_of_receipt;
				receive_payload_info[i].clean_up_after =
							time_of_receipt+2*(SEND_PERIOD+RECV_PERIOD);
				found_entry=TRUE;
			}
		}

		//if didn't find a pre-allocated entry, find an empty (or out of date) 
		//receive_payload_info table entry and fill it
		for (i=0; i<MAX_RCV_PLOAD_TABLE_SIZE && !found_entry; i++) {
			if (
				/* free */
				receive_payload_info[i].time_of_receipt==-1 || 
				/* or out of date */
				current_time > receive_payload_info[i].clean_up_after) {

				receive_payload_info[i].from = from;
				receive_payload_info[i].seq = seq;
				receive_payload_info[i].time_of_receipt = time_of_receipt;
				receive_payload_info[i].clean_up_after =
							time_of_receipt+2*(SEND_PERIOD+RECV_PERIOD);
				found_entry=TRUE;
			}
		}

		LeaveCriticalSection(&tcb_critical_section);

		//if an entry was filled, return TRUE; otherwise return FALSE
		if (found_entry) {
			srrt->response.service_id=MSG_RCV_ON_PAYLOAD;
			srrt->response.deliver=TRUE;
			srrt->type=2;
			//sinalize dll that answer is ready
			SetEvent(dll_wait_event);
		}
		else {
			//in this situation, msg is not delivered to application
			//to guarantee system correction
			srrt->response.service_id=MSG_RCV_ON_PAYLOAD;
			srrt->response.deliver=FALSE;
			srrt->type=2;
			//sinalize dll that answer is ready
			SetEvent(dll_wait_event);
		}

	}
	else {
		LeaveCriticalSection(&tcb_critical_section);
		//FD info was already received in the control channel.
		//All FD processing was already made.
		//fill answer
		srrt->response.service_id=MSG_RCV_ON_PAYLOAD;
		srrt->response.deliver=TRUE;
		srrt->type=2;
		//sinalize dll that answer is ready
		SetEvent(dll_wait_event);
	}
}



/**
 * THREAD: ServicesThread
 *
 * GOAL: to receive and dispatch service requests.
 *		 Replies are done by the dispatchers. 
 */
DWORD WINAPI ServicesThread (LPVOID param) {
	debug_tcb_services(("Waiting for requests...\n"));
	while(1) {
		//wait for a request
		WaitForSingleObject(tcb_wait_event, INFINITE);
		//after receiving the request, set event to false
		ResetEvent(tcb_wait_event);

		//confirm that there is a request waiting for a response
		if (srrt->type==1) {
			debug_tcb_services(("Answering service...\n"));
			switch (srrt->request.service_id) {
				case GET_TIMESTAMP:
					//call service
					getTimestamp();
					break;
				case GET_GLOBAL_TIMESTAMP:
					//call service
					getGlobalTimestamp(srrt->request.localInstant);
					break;
				case START_MEASUREMENT:
					//call service
					startMeasurement(srrt->request.start_ev);
					break;
				case END_MEASUREMENT:
					//call service
					endMeasurement(srrt->request.tag);
					break;
				case TIMELY_EXEC:
					//call service
					timelyExec(srrt->request.execArguments.start_ev, 
							   srrt->request.execArguments.delay, 
							   srrt->request.execArguments.max_exec, 
							   srrt->request.execArguments.dllName, 
							   srrt->request.execArguments.funcName, 
							   srrt->request.execArguments.wcetName);
					break;
				case START_LOCAL_FD:
					//call service
					startLocalFD(srrt->request.localFDArguments.start_ev, 
							     srrt->request.localFDArguments.spec, 
							     srrt->request.localFDArguments.deadline, 
							     srrt->request.localFDArguments.dllName, 
							     srrt->request.localFDArguments.funcName, 
							     srrt->request.localFDArguments.wcetName);
					break;
				case END_LOCAL_FD:
					//call service
					endLocalFD(srrt->request.tag);
					break;

				case GET_TCB_ID:
					//call service
					getTCBId();
					break;

				case START_DISTRIBUTED_FD:
					//call service
					startDistributedFD(
						srrt->request.distFDArguments.ec_id,
						srrt->request.distFDArguments.send_ev, 
						srrt->request.distFDArguments.spec, 
						srrt->request.distFDArguments.deadline, 
						srrt->request.distFDArguments.dllName, 
						srrt->request.distFDArguments.funcName, 
						srrt->request.distFDArguments.wcetName);
					break;
				
				case MSG_RCV_ON_PAYLOAD:
					//call service
					msgRcvOnPayload (
						srrt->request.msgRcvOnPayloadArguments.seq,
						srrt->request.msgRcvOnPayloadArguments.from,
						srrt->request.msgRcvOnPayloadArguments.time_of_receipt);
					break;

				case DEFINE_REMOTE_HANDLER:
					//call service
					defineRemoteFDHandler(
						srrt->request.remote_handler.dllName, 
						srrt->request.remote_handler.funcName, 
						srrt->request.remote_handler.wcetName,
						srrt->request.remote_handler.deadline);
					break;

				case START_DISTRIBUTED_REMOTE_FD:
					//call service
					startDistributedRemoteFD(
						srrt->request.distFDArguments.ec_id,
						srrt->request.distFDArguments.send_ev, 
						srrt->request.distFDArguments.spec);
					break;
				case SUBSCRIBE_EVENT_CHANNEL:
					//call service
					subscribeEventChannel(
						srrt->request.ec_id);
					break;
				case UNSUBSCRIBE_EVENT_CHANNEL:
					//call service
					unsubscribeEventChannel(
						srrt->request.ec_id);
					break;
				case MON_EVENT_CHANNEL:
					//call service
					monEventChannel(
						srrt->request.monEventChannelArguments.ec_id,
						srrt->request.monEventChannelArguments.mode,
						srrt->request.monEventChannelArguments.bound,
						srrt->request.monEventChannelArguments.coverage,
						srrt->request.monEventChannelArguments.deviation,
						srrt->request.monEventChannelArguments.num_samples,
						srrt->request.monEventChannelArguments.samples_source,
						srrt->request.monEventChannelArguments.dllName,
						srrt->request.monEventChannelArguments.funcName);
					break;
				case UPDATE_MON_EVENT_CHANNEL:
					//call service
					updateMonEventChannel(
						srrt->request.updateMonEventChannelArguments.tag,
						srrt->request.updateMonEventChannelArguments.update_ec_id,
						srrt->request.updateMonEventChannelArguments.ec_id,
						srrt->request.updateMonEventChannelArguments.update_mode,
						srrt->request.updateMonEventChannelArguments.mode,
						srrt->request.updateMonEventChannelArguments.update_bound,
						srrt->request.updateMonEventChannelArguments.bound,
						srrt->request.updateMonEventChannelArguments.update_coverage,
						srrt->request.updateMonEventChannelArguments.coverage,
						srrt->request.updateMonEventChannelArguments.update_deviation,
						srrt->request.updateMonEventChannelArguments.deviation,
						srrt->request.updateMonEventChannelArguments.update_num_samples,
						srrt->request.updateMonEventChannelArguments.num_samples,
						srrt->request.updateMonEventChannelArguments.update_samples_source,
						srrt->request.updateMonEventChannelArguments.samples_source);
					break;
				case END_MON_EVENT_CHANNEL:
					//call service
					endMonEventChannel(
						srrt->request.tag);
					break;
			}
		}
	}

	//never reaches here
	//return 0;
}


/************************************************************
 * TCB Core implementation
 ************************************************************/


/**
 * FUNCTION: push_send_queue
 *
 * GOAL: Inserts a msg with timestamp at the end of the send queue
 *
 * RETURNS: 0 if OK, -1 if queue is full
 *
int push_send_queue (data_msg_t dm, int timestamp) {
	//stores the error to be returned
	int error=0;

	//construct dmt to be inserted in the queue
	data_msg_with_timestamp_t dmt;
	dmt.dm=dm;
	dmt.timestamp=timestamp;

	EnterCriticalSection(&send_q);

	//if queue is empty
	if (empty_send) {
		first_send=0;
		last_send=0;
		empty_send=0;
		send_queue[0]=dmt;
	} 
	//if queue is full
	else if((last_send+1)%SIZE_SEND_QUEUE == first_send) {
		error=-1;
	}
	//insert at the end
	else {
		last_send = (last_send+1)%SIZE_SEND_QUEUE;
		send_queue[last_send]=dmt;
	}
	LeaveCriticalSection(&send_q);

	return error;
}


/**
 * FUNCTION: pop_send_queue
 *
 * GOAL: Removes the first msg with timestamp of the send queue and returns it
 *
 * RETURNS: 0 if OK, -1 if queue is empty
 *
int pop_send_queue (data_msg_t *dm, int *timestamp) {
	//stores the error to be returned
	int error=0;
	EnterCriticalSection(&send_q);
	//if queue is empty
	if (empty_send) {
		error=-1;
	} 
	//if queue has only one element
	else if (first_send==last_send) {
		*dm=send_queue[first_send].dm;
		*timestamp=send_queue[first_send].timestamp;
		first_send=-1;
		last_send=-1;
		empty_send=1;
	}
	//remove the first msg
	else {
		*dm=send_queue[first_send].dm;
		*timestamp=send_queue[first_send].timestamp;
		first_send=-1;
		first_send = (first_send+1)%SIZE_SEND_QUEUE;
	}

	LeaveCriticalSection(&send_q);
	return error;
}



/**
 * FUNCTION: print_recv_queue
 *
 * GOAL: Prints the content of the recv queue
 *
void print_recv_queue () {
	
	printf("Recv Queue Content:\n");
	printf("first=%d,last=%d,empty=%d\n",first_recv, last_recv, empty_recv);

	//if queue is empty
	if (empty_recv) {
		printf("Queue empty!\n");
	} 
	else {

		if (first_recv==-1 || last_recv==-1) {
			printf("FATAL error: (print)recv_queue should be empty!\n");
		}
		else
		 for (int i=0; i<SIZE_RECEIVE_QUEUE; i++)
			printf("src id=%d, seq=%d\n",recv_queue[i].dm.hdr.machine_src_id,
										 recv_queue[i].dm.hdr.seq);
	}
	
}



/**
 * FUNCTION: push_recv_queue
 *
 * GOAL: Inserts a msg with timestamp at the end of the recv queue
 *
 * RETURNS: 0 if OK, -1 if queue is full
 *
int push_recv_queue (data_msg_t dm, int timestamp) {
	//stores the error to be returned
	int error=0;
	//construct dmt to be inserted in the queue
	data_msg_with_timestamp_t dmt;
	dmt.dm=dm;
	dmt.timestamp=timestamp;

	EnterCriticalSection(&recv_q);

	//if queue is empty
	if (empty_recv) {
		first_recv=0;
		last_recv=0;
		empty_recv=0;
		recv_queue[0]=dmt;
	} 
	//if queue is full
	else if( ((last_recv+1)%SIZE_RECEIVE_QUEUE) == first_recv) {
		error=-1;
	}
	//insert at the end
	else {
		
		if (first_recv==-1 || last_recv==-1) {
			printf("FATAL error: (push)recv_queue should be empty!\n");
		}
		else {
			last_recv = (last_recv+1)%SIZE_RECEIVE_QUEUE;
			recv_queue[last_recv]=dmt;
			
#ifdef TCB_MEASUREMENTS
			
			//update counter if current size if greater than NUMBER_OF_MACHINES
			//(the normal value)
			int current_size=0;
			if (last_recv>=first_recv) {
				current_size = last_recv - first_recv +1;
			}
			else {
				current_size = (SIZE_RECEIVE_QUEUE-first_recv)+last_recv+1;
			}
			if (current_size>NUMBER_OF_MACHINES) {
				
				//printf("Warning: Recv Queue has more (%d) than %d msg\n",
				//				current_size,NUMBER_OF_MACHINES);
				num_rcv_queue_overflow++;
			}
#endif
		}
	}
	LeaveCriticalSection(&recv_q);

	return error;
}


/**
 * FUNCTION: pop_recv_queue
 *
 * GOAL: Removes the first msg with timestamp of the recv queue and returns it
 *
 * RETURNS: 0 if OK, -1 if queue is empty
 *
int pop_recv_queue (data_msg_t *dm, int *timestamp) {
	//stores the error to be returned
	int error=0;

	EnterCriticalSection(&recv_q);

	//if queue is empty
	if (empty_recv) {
		error=-1;
	} 
	//if queue has only one element
	else if (first_recv==last_recv) {
		
		if (first_recv==-1 || last_recv==-1) {
			printf("FATAL error: (pop)recv_queue should be empty!\n");
		}
		else {
			*dm=recv_queue[first_recv].dm;
			*timestamp = recv_queue[first_recv].timestamp;
			first_recv=-1;
			last_recv=-1;
			empty_recv=1;	
		}
	}
	//remove the first msg
	else {

		if (first_recv==-1 || last_recv==-1) {
			printf("FATAL error: (pop)recv_queue should be empty!\n");
		}
		else {
			*dm=recv_queue[first_recv].dm;
			*timestamp = recv_queue[first_recv].timestamp;
			first_recv = (first_recv+1)%SIZE_RECEIVE_QUEUE;
		}
	}
	LeaveCriticalSection(&recv_q);

	return error;
}
*/

//indicates when to send an id msg
bool sendIdMsg = TRUE;

/***
 * FUNCTION: send_to_all
 *
 * GOAL: Sends a message to all machines
 *
 * RETURNS: 0 if ok, -1 if error
 */
int send_to_all ( char* buf, int size_buf ) {

#ifdef EMUL_BROADCAST
   if (sendIdMsg) {
	//send id msg to all
	id_msg_t imt;
	imt.type = ID;
	imt.machine_src_id = my_id;
	struct sockaddr_in target;
	target.sin_addr.s_addr = inet_addr(EMUL_BROADCAST);
	target.sin_port = htons(TCB_PORT);
	target.sin_family = AF_INET;
	//debug_net(("Before sending to all..\n"));
	//send
	int nbytes = sendto(s, (char*)&imt, sizeof(id_msg_t), NULL, (struct sockaddr*)&target, 
						sizeof(target)); 
	//debug_net(("After sending to all..\n"));
	if (nbytes!=sizeof(id_msg_t))
		return -1;
   }
   //don't no if ~ is safe
   sendIdMsg==true?sendIdMsg=false:sendIdMsg=true;

	//send data msg pt2pt
	for (int i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		//send to others
		if (!membership[i].empty && membership[i].id!=my_id) {
			struct sockaddr_in target;
			int nbytes;
			target.sin_addr.s_addr = membership[i].id;
			target.sin_port = htons(TCB_PORT);
			target.sin_family = AF_INET;
			//send
			nbytes = sendto(s, buf, size_buf, NULL, (struct sockaddr*)&target, 
						sizeof(target)); 
			if (nbytes!=size_buf)
				return -1;
		}
		//copy to loopback buffer if msg is for me
		else if (!membership[i].empty) {
			
			struct sockaddr_in target;
			int nbytes;
			target.sin_addr.s_addr = membership[i].id; 
			target.sin_port = htons(TCB_PORT);
			target.sin_family = AF_INET;
			//send
			nbytes = sendto(s, buf, size_buf, NULL, (struct sockaddr*)&target, 
						sizeof(target)); 
			if (nbytes!=size_buf)
				return -1;

			/*
			EnterCriticalSection(&loopback_critical_section);
			//if buffer is empty
			if (loop_first==-1 && loop_last==-1) {
				loop_first = loop_last = 0;
			}
			else {
				//if buffer is full
				if (loop_first == (loop_last+1)%MAX_SIZE_LOOPBACK_BUFFER){
					printf("FATAL ERROR: Loopback buffer is full!\n");
				}
				else {
					loop_last = (loop_last+1)%MAX_SIZE_LOOPBACK_BUFFER;
				}
			}
			//add info to buffer last position
			//printf("Added msg to loop pos %d\n", loop_last);
			memcpy(loopbt[loop_last].buffer, buf, size_buf);
			loopbt[loop_last].size = size_buf;
			
			LeaveCriticalSection(&loopback_critical_section);
			*/
		}
	}
#endif
#ifdef REAL_BROADCAST
	struct sockaddr_in target;
	target.sin_addr.s_addr = inet_addr(REAL_BROADCAST);
	//if real broadcast is being used, 
	//all machines are binded in the same port
	target.sin_port = htons(TCB_PORT);
	target.sin_family = AF_INET;
	//debug_net(("Before sending to all..\n"));
	//send
	int nbytes = sendto(s, buf, size_buf, NULL, (struct sockaddr*)&target, 
						sizeof(target)); 
	//debug_net(("After sending to all..\n"));
	if (nbytes!=size_buf)
		return -1;
#endif
	return 0;
}


/***
 * FUNCTION: send_to
 *
 * GOAL: Sends a message to a machine
 *
 * RETURNS: 0 if ok, -1 if error
 *
int send_to (char id, char* buf, int size_buf) {
	struct sockaddr_in target;
	int nbytes;
	target.sin_addr.s_addr = inet_addr(machines[id].ip);
	target.sin_port = htons(machines[id].port);
	target.sin_family = AF_INET;
	//send
	nbytes = sendto(s, buf, size_buf, NULL, (struct sockaddr*)&target, 
					sizeof(target)); 
	if (nbytes!=size_buf)
		return -1;
	return 0;
}
*/

/**
 * FUNCTION: send_token
 *
 * GOAL: Sends a token to our successor in the logic ring
 *
 * RETURNS: 0 if OK, -1 if error
 *
int send_token () {
	int error=0;
	token_msg_t tm;
	tm.type=TOKEN;
	tm.machine_src_id=my_id;
	tm.seq=++token_seq_num;
	//debug_net(("Sending token seq=%d\n", tm.n_seq));
	for (int i=TOKEN_OMISSION_DEGREE+1; i--;) {
		//send	
		error=send_to(successor, (char*)&tm, sizeof(tm));
	}
	return error;
}
*/


/**
 * FUNCTION: buildRoundTripInfo
 *
 * GOAL: Fills a data msg with the roundtrip info, including snd_time
 */
int buildRoundTripInfo (data_msg_t* dm) {
	 int now = GetTickCount();

	 for (int i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		 if (membership[i].empty) {
			 dm->hdr.RoundTripInfo[i].empty = TRUE;	 
			 continue;
		 }
		 dm->hdr.RoundTripInfo[i].empty = FALSE;	 
		 dm->hdr.RoundTripInfo[i].id = membership[i].id;
		 dm->hdr.RoundTripInfo[i].snd_time = membership[i].last_rcv_info.snd_time;
		 dm->hdr.RoundTripInfo[i].rndval = now - membership[i].last_rcv_info.timestamp;
	 }
	 dm->hdr.snd_time=now;
	 return now;
 }


/**
 * FUNCTION: updateInfoFields
 *
 * GOAL: Updates info fields to provide better accuracy in the FD
 *
 * PARAM:
 *		  dm - Data msg with the info_fields
 *		  diff - Milliseconds since creation of info_fields
 */
void updateInfoFields (data_msg_t* dm, int diff) {
	 info_field_t *info_ptr = (info_field_t*)dm->data;

	 for (int i=0; i<dm->hdr.num_info_fields; i++) {
		//update only requests
		if (info_ptr->type==0)
			info_ptr->time_to_failure -= diff;
		info_ptr++;
	 }
 }


/**
 * THREAD: SendToNetworkThread
 *
 * GOAL: to send msgs to network.
 *		 Waked by NetworkThread when there is a data msg to send.
 *
DWORD WINAPI SendToNetworkThread (LPVOID param) {
	
	while(1) {
		//suspend myself until a wakeup from SendThread
		SuspendThread(hSendToNet);

		
	}
}
*/



/**
 * THREAD: NetworkThread
 *
 * GOAL: to send/receive msgs to/from network according to 
 *		 the "Ad-Hoc Token Ring protocol"
 *
DWORD WINAPI NetworkThread (LPVOID param) {

	int error=0;
	data_msg_t* dm_ptr=NULL;
	token_msg_t* tm_ptr=NULL;
	data_msg_t dm;
	token_msg_t tm;
	//to store sequence number and id of msg recved
	int n_seq, id;
	//pointer to last seq number recved 
	//(token or data) of msg sender
	int* last_seq_num=NULL;
	//indicates if we are going to recv first token
	BOOL first_token=TRUE;


#ifdef TCB_MEASUREMENTS
	for (int i=0; i<NUMBER_OF_MACHINES; i++) {
		total_roundtripVal[i]=0;
		num_roundtripVal[i]=0;
		min_roundtripVal[i]=10000;
		max_roundtripVal[i]=0;
	}
#endif



	//if we are machine 0, send token to initialize the ring
	if (my_id==0) {
		//debug_net(("Sending init token...\n"));
		send_token();
	}
	
	//store current timestamp to measure how much time
	//TCB runs until a timeout occurs
	int TCBstartTime = GetTickCount();

	//main loop
	while(1){
		//sleep NET_PERIOD ms to let other applications run
		//(this is only necessary when we have only 1 TCB)
#if NUMBER_OF_MACHINES==1
		Sleep(NET_PERIOD);
#endif
		//debug_net(("Net:After sleeping...\n"));

		//only do select if we have already received the first token
		//Note: first_token is set to FALSE after awakening Services thread
		if (((!first_token) || (first_token && my_id==0)) && NUMBER_OF_MACHINES>1) {
			//wait msg (token or data) or timeout
			fd_set set;
			FD_ZERO(&set);
			FD_SET(s, &set);
			timeval tv;
			tv.tv_sec=MSG_TIMEOUT/1000;
			tv.tv_usec=(MSG_TIMEOUT%1000)*1000;
			//debug_net(("Net:On select...\n"));
			error=select(1, &set, NULL, NULL, &tv);
			if (error==0) {
				printf("FATAL error: Timeout on receiving msg\n");
				int TCBtimeElapsed = GetTickCount() - TCBstartTime;
				printf("TCB runned %d seconds\n", TCBtimeElapsed/1000);
#ifdef TCB_MEASUREMENTS
				for (i=0; i<NUMBER_OF_MACHINES; i++) {
					//prints average roundtrip
					printf("%d msg,(AV/MIN/MAX)delivery delay from %d=(%.2f/%d/%d) %d om.\n",
							num_roundtripVal[i], i, 
							((double)total_roundtripVal[i])/num_roundtripVal[i],
							min_roundtripVal[i],max_roundtripVal[i], n_omissions);
				}
				//prints average time between producing and sending a data msg
				printf("send_q latency=(%.2f/%d/%d), rcv_q latency=%d,num_rcv_queue_overflow=%d\n", 
								((double)total_in_send_queue_latency)/num_in_send_queue_latency,
								min_in_send_queue_latency,
								max_in_send_queue_latency, 
								max_before_recv_queue_latency,
								num_rcv_queue_overflow);

				//prints average time that a token takes to circle by all TCBs
				printf("%d msg,(AV/MIN/MAX)token delay:(%.2f/%d/%d)\n",
							num_token_delay, 
							((double)total_token_delay)/num_token_delay,
							min_token_delay,max_token_delay);

#ifdef USE_ACK
				//prints average number of retransmits
				printf("(AV/MIN/MAX)retransmits:(%.2f/%d/%d)\n",
							((double)total_retransmits)/num_retransmits,
							min_retransmits,max_retransmits);
#endif

#ifdef LOG_QUEUES

				// print wait_info content for debug
				printf("Wait info content at %d:\n", GetTickCount());
				for (i=0; i<MAX_WAIT_INFO_TABLE_SIZE; i++) {
					if (wait_info[i].failure_at_time!=-1) {
						printf("from=%d,seq=%d,failure_at_time=%d,n_dests=%d, destMask=%d\n",
								wait_info[i].from, wait_info[i].seq,
								wait_info[i].failure_at_time,
								wait_info[i].n_dests,
								wait_info[i].destMask);
					}
				}


				
				printf("\nProcRequest queue log: %d entries\n", procRequest_log_size);
				for (i=0; i<procRequest_log_size; i++) {
					printf("from=%d, seq=%d\n", procRequest_log[i].from, procRequest_log[i].seq);
				}
				
				


#endif


#endif
				//prints number of data msg that were lost
				printf("#omissions=%d\n", n_omissions);
				printf("Last token sent=%d\n", token_seq_num);
				printf("Last data sent=%d\n", data_seq_num);
				printf("Last token recved from %d=%d\n",
								predecessor,last_token_seq_num_recv);
				for (int i=0; i<NUMBER_OF_MACHINES; i++) {
					printf("Last data recved from %d=%d\n",
								i,last_rcv_info[i].seq);
				}
				break;
			}
		}

		
			// if it reaches here, msg can be processed 
			//if it's a token, transmit first msg in the send queue
			if (buffer[0]==TOKEN) {
				//if it's the first token received, unblock Services thread
				if (first_token) {
					SetEvent(tcb_network_sync);
					first_token=FALSE;
					//debug_net(("Net:Unblocking TCB threads...\n"));
				}

#ifdef TCB_MEASUREMENTS
					//measures how much time the token takes to go around and return
					int now2 = GetTickCount();
					if (last_token_instant!=-1) {
						int token_delay = now2-last_token_instant;
						total_token_delay += token_delay;
						num_token_delay++;
						if (token_delay>max_token_delay)
							max_token_delay=token_delay;
						if (token_delay<min_token_delay)
							min_token_delay=token_delay;
						
					}
					last_token_instant = now2;
#endif
				//only send 1 data msg per TOKENS_PER_SEND tokens.
				if ((n_seq%TOKENS_PER_SEND) != 0) {
					send_token();
					//update last sequence number received
					*last_seq_num=n_seq;
					continue;
				}
				
				debug_net(("Net:Token recved\n"));
				//stores the timestamp used to construct the info fields
				int info_fields_timestamp;
				error=pop_send_queue(&dm_global, &info_fields_timestamp);
				
				//if queue is empty, don't transmit anything
				if (error==-1) {
					//resume SendThread to produce one more message
					ResumeThread(hSend);
					//debug_net(("Send queue empty: not sending\n"));
					send_token();
				}
				//send the popped data msg
				else {
					
				} 
			}
			//if it's a data msg
			else {
				//verify that we received the correct number of bytes
				int num_info_fields = dm_ptr->hdr.num_info_fields;
				if (nbytes != sizeof(data_msg_hdr)+
								num_info_fields*sizeof(info_field_t)) {

					printf("Error: Msg size is wrong\n");
				}
				else {	
					
					
#ifdef TCB_MEASUREMENTS
					//to calculate the maximum time since a msg being received and
					//pushed in the recv_queue
					int latency = GetTickCount()-t;
					if (latency>max_before_recv_queue_latency)
						max_before_recv_queue_latency=latency;
#endif







	//close socket
	closesocket(s);
	return 0;
}
*/


int get_wait_info_pos (int seq, unsigned long from);

 
 /**
  * FUNCTION: buildDataInfo
  *
  * GOAL: 1) Fills the info fields of a data msg, with:
  *
  *			a)the content of send_info table (FD requests)
  *
  *			b)the FD results of receive_info table already known
  *			  (FD responses)
  *
  *	PARAM:
  *		  dm - Data msg to be filled
  *		  ts - Data msg send time instant
  */
 void buildDataInfo (data_msg_t* dm, int ts) {
	 
	 //debug_tcb(("builDataInfo(ts=%d)\n",ts));

	 //stores the current info field to be filled
	 info_field_t *info_ptr = (info_field_t*) dm->data;
	 //store the total number of info_fields filled
	 int num_info_fields = 0;

	 //scan send_info table for pending sends
	 for (int i=0; i<MAX_SEND_INFO_TABLE_SIZE && 
					num_info_fields<MAX_NUM_INFO_FIELDS; i++) {
		 if (send_info[i].time_of_generation != -1) {
			 debug_tcb(("(buildDataInfo) process pend send\n"));
			 int ms_since_generation = ts-send_info[i].time_of_generation;
			 info_ptr->time_to_failure = send_info[i].spec-ms_since_generation;
			 memcpy(info_ptr->destination, send_info[i].destination, MAX_NUMBER_OF_MACHINES*sizeof(unsigned long));
			 info_ptr->seq = send_info[i].seq;
			 info_ptr->remote_fd = send_info[i].remote_fd;
			 info_ptr->ec_id = send_info[i].ec_id;
			 info_ptr->spec = send_info[i].spec;
			 //it's a request
			 info_ptr->type = 0;
			 num_info_fields++;
			 //point to next info_field
			 info_ptr++;
			 //free send_info position
			 send_info[i].time_of_generation = -1;
		 }
	 }

	 //scan receive_info table for known FD results
	 for (i=0; i<MAX_RCV_INFO_TABLE_SIZE && 
		 num_info_fields<MAX_NUM_INFO_FIELDS; i++) {
		 //not empty
		 if (receive_info[i].failure_at_time != -1) {
			//FD result is known if:
			//1) failure instant has been reached and a timing failure occured
			//2) msg was timely received
			if ((ts>=receive_info[i].failure_at_time &&
				receive_info[i].fail==1) || receive_info[i].fail==0) {
				
				debug_tcb(("(buildDataInfo) process known result\n"));
	
				//If it's a remote FD and a failure occurred
				if (receive_info[i].remote_fd && receive_info[i].fail==1) {
					/*call fail handler*/
					//get position of wait_info table entry that stores FD operation status
					int wait_info_pos = get_wait_info_pos(receive_info[i].seq,
														  receive_info[i].from);
					//if entry is not found, it's not an error because check_blocked_tfd must
					//have runned before
					if (wait_info_pos==-1) {
						
						//printf("FATAL ERROR: Couldn't find wait_info position for calling fail handler in a remote FD, from=%u, seq=%d\n",
						//			receive_info[i].from, receive_info[i].seq);
					}
					else {
						if (wait_info[wait_info_pos].task_id != NULL) {
							ResumeThread(wait_info[wait_info_pos].task_id);
						}
					}
				}

				info_ptr->fail = receive_info[i].fail;
				info_ptr->lateness_avail = receive_info[i].lateness_avail;
				if (info_ptr->lateness_avail) {
					info_ptr->lateness_degree = receive_info[i].lateness_degree;
				}
				else info_ptr->lateness_degree = 0;
				info_ptr->seq = receive_info[i].seq;
				info_ptr->destination[0] = receive_info[i].from;

				//if msg was sent only for me, then type=1 to avoid that
				//other TCBs process this field
				//char my_mask = 1<<my_id;
				//if (receive_info[i].destMask==my_mask) {
				//	info_ptr->type =1;
				//}
				info_ptr->type =2;
			
				num_info_fields++;
				info_ptr++;
				//free receive_info position
				receive_info[i].failure_at_time = -1;
			}
		 }
	 }

	 //indicate how much info_fields were filled
	 dm->hdr.num_info_fields = (char)num_info_fields;
 }


 

/**
 * THREAD: SendThread
 *
 * GOAL: to periodically send a msg to all TCBs
 */
DWORD WINAPI SendThread (LPVOID param) {
	//printf("SendThread running\n");
	data_msg_t dm;
	
	while(1){		
		EnterCriticalSection(&tcb_critical_section);

		int now = GetTickCount();
		//fill info fields
		buildDataInfo(&dm, now);
		//reset number of sends for the following send period
		n_sends_in_this_slot = 0;
		
		//fill type
		dm.hdr.type=DATA;
		//fill our id
		dm.hdr.machine_src_id=my_id;
		//fill sequence number
		dm.hdr.seq = ++data_seq_num;
		//fill ec_id
		tcb_info_t* ti = getTCB(my_id);
		if (ti==NULL) {
			printf("Error: I doesn't exist!\n");
			return -1;
		}
		dm.hdr.ec_id = ti->ec_ids;
		
		debug_net(("Sending data seq=%d\n", dm.hdr.seq));
				
		//fill with roundtrip information and return snd_time used
		int now2 = buildRoundTripInfo(&dm);
		//update info fields to provide a better accuracy of the
		//failure detector.
		updateInfoFields(&dm,now2-now);

		LeaveCriticalSection(&tcb_critical_section);

		//fill with current global clock value
		dm.hdr.vclock = get_vclock_at(now2);
									
		//send msg DATA_OMISSION_DEGREE+1 times
		//(we only send the number of info fields specified)
	    for (int i=DATA_OMISSION_DEGREE+1; i--;) {
			if (send_to_all((char*)&dm, 
						sizeof(data_msg_hdr_t)+
						dm.hdr.num_info_fields*
						sizeof(info_field_t)
						)==-1)
			
				printf("Error sending msg to all machines\n");
			
		}

		int elapsed = GetTickCount()-now;
		
		if (SEND_PERIOD-elapsed>0)
			Sleep(SEND_PERIOD-elapsed);
		
		else
			printf("Warning: SendThread runned during %d ms and SEND_PERIOD=%d\n",
								elapsed, SEND_PERIOD);
		
	}
	//never reaches here
	//return 0;
}


/**
 * FUNCTION: checkRoundTrip
 *
 * GOAL: Calculates maximum roundtrip value of a received msg
 *
 * PARAM:
 *		dm Data msg received
 *		timestamp Time of receipt
 *
 * RETURNS: maximum roundtrip value, -1 if error
 */
int checkRoundTrip (data_msg_t *dm, int timestamp) {
	
	//debug_tcb(("checkRtrip:Using src_id=%d\n",dm->hdr.machine_src_id));
	
	//calculate the upper limit of the delivery time value.
	
	//(calculate only if snd_time of our last msg is >0, i.e. 
	//if the sender already received a msg of us)
	int index = -1;
	for (int i=0; i<MAX_NUMBER_OF_MACHINES && index==-1; i++) {
		if (!dm->hdr.RoundTripInfo[i].empty &&
			 dm->hdr.RoundTripInfo[i].id == my_id)
			 index = i;
	}
	if (index==-1) {
		debug_tcb(("Warning: no round trip info. Should be first msg\n"));
		return MAXIMUM_DELIVERY_DELAY;
	}
	
	if (dm->hdr.RoundTripInfo[index].snd_time<=0) {
		debug_tcb(("Warning: snd_time<=0. Should be first msg\n"));
		return MAXIMUM_DELIVERY_DELAY;
	}

	int roundtripVal = timestamp - dm->hdr.RoundTripInfo[index].snd_time;
	roundtripVal -= dm->hdr.RoundTripInfo[index].rndval;
		
	//Due to the fact that full-duplex is done by software, it's not considered
	//the minimum delivery delay for msgs sent and received by the same machine.
	if (dm->hdr.machine_src_id!=my_id) {
		roundtripVal -= MINIMUM_DELIVERY_DELAY;
	}
		
	if (roundtripVal>MAXIMUM_DELIVERY_DELAY) {
		//printf("FATAL error: MAXIMUM_DELIVERY_DELAY exceeded: %d\n", roundtripVal);
		//return -1;
		
		//for printing IP of source TCB
		struct in_addr in;
		in.S_un.S_addr = dm->hdr.machine_src_id;
		printf("Warning: MAXIMUM_DELIVERY_DELAY exceeded: %d ms from %s\n", roundtripVal, 
				inet_ntoa(in));
		//return -1;
	}
	
	//roundTripVal can be negative if the msg has a very low delivery time value
	if (roundtripVal<0) {
		/*
		printf("Warning: rndtripVal<0 from %d, type=%d, because:timestamp=%d,last_snd_time=%d,rndval=%d,sndTime=%d\n",
						dm->hdr.machine_src_id, dm->hdr.type, 
						timestamp, dm->hdr.RoundTripInfo[my_id].snd_time,
						dm->hdr.RoundTripInfo[my_id].rndval,
						dm->hdr.snd_time);
		*/
		roundtripVal=0;
	}
		
	return roundtripVal;
}


/**
 * FUNCTION: already_received_on_payload
 *
 * GOAL: Verifies if it was already received a msg from (from) with seq=(seq)
 *	     in the payload channel. If yes, checks if it was timely or not.
 *		 Called by processRequest.
 *
 * PARAM:
 *	rcv_info_index - index of the receive_info entry used by the caller function
 *	from_ti - info about sender TCB
 *	seq - sequence number that identifies FD operation
 *	failure_at_time - failure instant
 */
void already_received_on_payload (int rcv_info_index, tcb_info_t *from_ti, int seq, 
								  int failure_at_time) {
	
	debug_tcb(("already_rcv_pload(from=%d,seq=%d)\n",from_ti->id, seq));

	//scan receive payload info table to discover if already received
	//a msg from (from) and with seq (seq)
	for (int i=0; i<MAX_RCV_PLOAD_TABLE_SIZE; i++) {
		if (receive_payload_info[i].time_of_receipt != -1 &&
			receive_payload_info[i].time_of_receipt != -2 &&
			receive_payload_info[i].from == from_ti->id &&
			receive_payload_info[i].seq == seq) {
			
			/* check if it was timely */
			if (receive_payload_info[i].time_of_receipt > failure_at_time) {
				//timing failure!
				receive_info[rcv_info_index].fail = 1;
				receive_info[rcv_info_index].lateness_avail = true;
				receive_info[rcv_info_index].lateness_degree = 
					receive_payload_info[i].time_of_receipt - failure_at_time;
			}
			else {
				//timely!
				receive_info[rcv_info_index].fail = 0;
				receive_info[rcv_info_index].lateness_avail = true;
				receive_info[rcv_info_index].lateness_degree = 
					 failure_at_time - receive_payload_info[i].time_of_receipt;
			}
			//clean entry
			receive_payload_info[i].time_of_receipt = -1;
			return;
		}
	}
}


/**
 * FUNCTION: existsDestination
 *
 * GOAL: to verify if a tcb id is in a destination array
 *
 * RETURN: TRUE if it is, FALSE otherwise
 */
bool existsDestination(unsigned long tcb_id, unsigned long *destination) {
	for (int i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		if (destination[i] == tcb_id)
			return TRUE;
	}
	return FALSE;
}


/**
 * FUNCTION: processRequest
 *
 * GOAL: processes a distributed timing failure detection request.
 *		 Called by processRead.
 *
 * PARAMS:
 *		from_ti - info about sender tcb
 *		destination - destinations ids
 *		seq - sequence number that identifies FD operation
 *		time_to_failure - remaining milliseconds to failure when
 *						  msg was sent
 *		roundtripVal - maximum msg delivery time
 *		time_of_receipt - reception timestamp
 *		remote_fd - indicates if fail handler should be executed in recipients
 *      ec_id - event channel id
 *      spec - original specification of failure detection
 */
void processRequest (tcb_info_t* from_ti, unsigned long *destination, 
					 int seq, int time_to_failure, 
					 int roundtripVal, int time_of_receipt, bool remote_fd, 
					 char ec_id, int spec) {
	
	debug_tcb(("processRequest(from=%d,seq=%d)\n",from_ti->id,seq));

#ifdef LOG_QUEUES

	//if there is a <from,seq-1>, reuse that entry
	bool found=FALSE;
	for (int k=0; k<procRequest_log_size && !found; k++) {
		if (procRequest_log[k].from == from &&
				procRequest_log[k].seq == seq-1) {
			
					found = TRUE;
					procRequest_log[k].seq = seq;
		}
	}	
			
	if (!found) {
		procRequest_log[procRequest_log_size].from = from;
		procRequest_log[procRequest_log_size++].seq = seq;
	}

#endif

	bool to_array[MAX_NUMBER_OF_MACHINES];		
	char number_of_dest=0;
	bool am_i_subscribed = FALSE;

	//populate to_array and calculate total number of destinations
	for (int i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		if (membership[i].empty)
			continue;
		
		//if machine i is a destination
		if (existsDestination(membership[i].id, destination)) {
			to_array[i] = TRUE;
			number_of_dest++;
			if (membership[i].id==my_id)
				am_i_subscribed = TRUE;
		}
		else
			to_array[i]=FALSE;
	}

    //if msg is not for me, don't process it
	if (!am_i_subscribed) {
		/*
		if (from!=my_id) {
			printf("Warning: Msg from other TCB (%d) was not addressed to me\n",from);
		}
		*/
		return;
	}
	
	//find a empty entry of receive_info table and fill it
	for (i=0; i<MAX_RCV_INFO_TABLE_SIZE; i++) {
		if (receive_info[i].failure_at_time==-1) {
			receive_info[i].failure_at_time = 
				time_of_receipt+(time_to_failure-roundtripVal);
			memcpy(receive_info[i].destination, destination, MAX_NUMBER_OF_MACHINES*sizeof(unsigned long));;
			receive_info[i].from = from_ti->id;
			receive_info[i].seq = seq;
			receive_info[i].remote_fd = remote_fd;

		/*
#ifdef LOG_QUEUES
			//if there is a <from,seq-1>, reuse that entry
			bool found=FALSE;
			for (int k=0; k<rcv_log_size && !found; k++) {
				if (rcv_log[k].from == from &&
					rcv_log[k].seq == seq-1) {
				
					found = TRUE;
					rcv_log[k].seq = seq;
				}

			}
			
			if (!found) {
				rcv_log[rcv_log_size].from = from;
				rcv_log[rcv_log_size++].seq = seq;
			}
#endif
*/

			//by default, a failure occurs. If the msg is timely received
			//in the payload channel, this flag is changed to 0.
			//If the msg is not received in the payload channel until 
			//"failure_at_time" instant, then
			//this field already contains the right value.
			receive_info[i].fail = 1;
			receive_info[i].lateness_avail = false;
			receive_info[i].lateness_degree = 0;
			//verify if msg was already received in the payload. If yes,
			//check if it was timely.
			already_received_on_payload(i,from_ti,seq,
										receive_info[i].failure_at_time);

			HANDLE task_id = NULL;
			//if it's a remote FD and msg was not yet received on payload
			if (remote_fd && receive_info[i].fail) {
				//create task that will be responsible for executing fail handler
				//if a failure occurs.
			
				task_id = NULL;
				//if a failure handler was specified, init failTask
				if (remote_fd_handler!=NULL) {
				 /* prepare arguments for failTask */
				 //dfi is a global variable
				 dfi.funcptr = remote_fd_handler;
				 dfi.deadline = remote_fd_deadline;
				 dfi.from = from_ti->id;
				 dfi.seq = seq;
				 //for now, this value is not used
				 dfi.release_time=0;
				 dfi.failure_at_time = receive_info[i].failure_at_time;
				 //create thread that will monitor timing failure
				 task_id =
					CreateThread(NULL,0,DistFDThread,NULL,CREATE_SUSPENDED,NULL);
				 if (task_id==NULL){
					_wperror(L"FATAL ERROR: Couldn't start fail task for a remote FD.");
					return;
				 }
				 //set thread priority to MAXIMUM-2
				 CeSetThreadPriority(task_id, 2);
				 //thread is only started if a failure occurs

				 //verify if a failure has already occured
				 if (dfi.failure_at_time<GetTickCount()) {
					//start FailTask
					ResumeThread(task_id);
					//to avoid failTask being called more times
					task_id = NULL;
				 }
				} //endif(remote_fd_handler!=NULL)
				
			
				//verify if sender is also a recipient. In that case,
				//fill field task id.
				if (task_id!=NULL) {
					for (int j=0; j<MAX_WAIT_INFO_TABLE_SIZE; j++) {
						if (wait_info[j].failure_at_time != -1 &&
							wait_info[j].from == from_ti->id) {
						
							wait_info[j].task_id = task_id;
						}
					}
				}

			} //endif(remote_fd)

			
			/* find a empty entry of wait_info table and fill it.
			   this entry is later used to return FD result to application */
			//if we were the sender of the msg, we already filled a 
			//wait_info entry
			if (from_ti->id==my_id)
				return;

			for (int j=0; j<MAX_WAIT_INFO_TABLE_SIZE; j++) {
				if (wait_info[j].failure_at_time == -1) {

					wait_info[j].failure_at_time = 
										receive_info[i].failure_at_time;
					wait_info[j].any_fail = FALSE;
					memcpy(wait_info[j].destination, destination, MAX_NUMBER_OF_MACHINES*sizeof(unsigned long));
					wait_info[j].from = from_ti->id;
					wait_info[j].n_dests = number_of_dest;
					wait_info[j].seq = seq;
					//it is only used by sender (if remote_fd is FALSE)
					wait_info[j].task_id = NULL;
					if (remote_fd) {
						wait_info[j].task_id = task_id;
					}
					wait_info[j].remote_fd = remote_fd;
					wait_info[j].ec_id = ec_id;
					wait_info[j].spec = spec;
					//at this time, we don't have any failure info
					for (int k=0; k<MAX_NUMBER_OF_MACHINES; k++) {
						if (to_array[k]) {
							wait_info[j].failure_info[k].empty=FALSE;
							wait_info[j].failure_info[k].id = membership[k].id;
							//2=no info
							wait_info[j].failure_info[k].fail=2;
						}
						else
							wait_info[j].failure_info[k].empty=TRUE;
					}
/*
#ifdef LOG_QUEUES
					
					//if there is a <from,seq-1>, reuse that entry
					bool found=FALSE;
					for (k=0; k<wait_log_size && !found; k++) {
						if (wait_log[k].from == from &&
							wait_log[k].seq == seq-1) {
				
							found = TRUE;
							wait_log[k].seq = seq;
						}

					}	
			
					if (!found) {
						wait_log[wait_log_size].from = from;
						wait_log[wait_log_size++].seq = seq;
					}
					
#endif
*/
					return;
				}
			} //endfor (j)
			//if the system is correct, never reaches here
			printf("FATAL ERROR(proc_request): No free wait info entry\n");
		}
	} //endfor (i)

	//if the system is correct, never reaches here
	printf("FATAL ERROR(proc_request): No free receive info entry\n");
	wait_user();
}



/**
 * FUNCTION: process_qos_monitoring
 *
 * GOAL: to verify if we are monitoring the event channel associated
 *		 with the wait_info received and to process the new info
 *
 * PARAM:
 *		wait_info_pos index of entry in wait_info_table
 */
 void process_qos_monitoring (int wait_info_pos) {

	 //printf("process_qos_monitoring()\n");

	//verify if we are monitoring this event channel
	//(there can various monitorings of an event channel) 
	for (int t=0; t<MAX_QOS_MONITOR_INFO_TABLE_SIZE ; t++) {
		if (!qos_monitor_info[t].empty && 
			qos_monitor_info[t].ec_id==wait_info[wait_info_pos].ec_id) {
			
			qos_monitor_info_t *qmi = &qos_monitor_info[t];
	
			//verify if samples_source allows to process this FD info
			if (!(qmi->samples_source==0 ||
				(qmi->samples_source==1 && wait_info[wait_info_pos].from==my_id) ||
				(qmi->samples_source==2 && wait_info[wait_info_pos].from!=my_id)))
				continue;

			int new_samples[MAX_NUMBER_OF_MACHINES];
			for (int k=MAX_NUMBER_OF_MACHINES; k--;)
				new_samples[k]=-1;
			int j=0;
			//fill new_samples
			for (k=0; k<MAX_NUMBER_OF_MACHINES; k++) {
				if (!wait_info[wait_info_pos].failure_info[k].empty) {
				 //if timely, value=spec-lateness_degree
				 if (!wait_info[wait_info_pos].failure_info[k].fail)
					new_samples[j++]= wait_info[wait_info_pos].spec -	
						wait_info[wait_info_pos].failure_info[k].lateness_degree;
				 //if failed and lateness is available, value=spec+lateness_degree
				 else if (wait_info[wait_info_pos].failure_info[k].fail && 
						  wait_info[wait_info_pos].failure_info[k].lateness_avail)
						new_samples[j++]= wait_info[wait_info_pos].spec +
							wait_info[wait_info_pos].failure_info[k].lateness_degree;
				 //if failed and lateness is not available, value=MAX_DELIVERY_DELAY
				 else if (wait_info[wait_info_pos].failure_info[k].fail && 
						  !wait_info[wait_info_pos].failure_info[k].lateness_avail)
						new_samples[j++]= MAXIMUM_DELIVERY_DELAY;
				}
			} //end for

			//if there are no new samples
			if (j==0)
				continue;
	
			//process new_samples
			for (int b=0; b<j; b++) {

				//printf("Processing sample %d\n", new_samples[b]);

				int deleting_sample=-1;
				//insert in PDF samples history
				if (qmi->pdf.first_sample==-1) {
					/*when history is empty*/
					//init pointers (last_sample is inited below)
					qmi->pdf.first_sample = 0;
				} 
				else if (qmi->pdf.first_sample==(qmi->pdf.last_sample+1)%qmi->num_samples){
					/*when history is full*/
					//store and delete oldest sample
					deleting_sample = qmi->pdf.samples_history[qmi->pdf.first_sample];
					qmi->pdf.first_sample = (qmi->pdf.first_sample+1)%qmi->num_samples;
				}
				qmi->pdf.last_sample = (qmi->pdf.last_sample+1)%qmi->num_samples;
				//insert new sample
				qmi->pdf.samples_history[qmi->pdf.last_sample]=new_samples[b];
				//update mean
				if (deleting_sample!=-1) {
					//printf("Deleting sample...\n");
					if (qmi->pdf.num_samples==1)
						qmi->pdf.mean=0;
					else
						qmi->pdf.mean = ((qmi->pdf.mean*qmi->pdf.num_samples)-deleting_sample)/
							          (qmi->pdf.num_samples-1);
					qmi->pdf.num_samples--;
				}
				
				qmi->pdf.mean = (qmi->pdf.mean*qmi->pdf.num_samples+new_samples[b])/
								 (qmi->pdf.num_samples+1);
				qmi->pdf.num_samples++;

				//update variance
				double quadratic_sum=0;
				for (int a=qmi->pdf.first_sample,k=0; k<qmi->pdf.num_samples; a=(a+1)%qmi->num_samples,k++) {
					//printf("Adding sample %d to quadratic_sum\n", qmi->pdf.samples_history[a]);
					quadratic_sum += (qmi->pdf.samples_history[a]-qmi->pdf.mean)*
									 (qmi->pdf.samples_history[a]-qmi->pdf.mean);
				}
				//printf("Quadratic sum=%.2f\n", quadratic_sum);
				if (qmi->pdf.num_samples==1)
					qmi->pdf.variance = quadratic_sum;
				else
					qmi->pdf.variance = quadratic_sum/(qmi->pdf.num_samples-1);

				//printf("Mean/Variance=%.2f/%.2f\n", qmi->pdf.mean, qmi->pdf.variance);
			}

			//verify bound/coverage according to mode
			if (qmi->mode==0) {
				/*coverage stability*/
				//calculate bound current coverage
				double current_coverage=0;
				if (qmi->bound<=qmi->pdf.mean)
					current_coverage=0;
				else
					current_coverage = (1-(qmi->pdf.variance/(qmi->pdf.variance+
										((qmi->bound-qmi->pdf.mean)*
										(qmi->bound-qmi->pdf.mean)))));

				//printf("Current/Specified coverage: %.2f/%.2f\n", current_coverage, qmi->coverage);

				//if current coverage is deviated more than specified, 
				//call handler with the new bound that verifies the specified coverage
				if (fabs(qmi->coverage-current_coverage)>qmi->deviation) {
					int new_bound = qmi->pdf.mean+sqrt( (qmi->pdf.variance/(1-qmi->coverage))
								    - qmi->pdf.variance);
					qmi->handler(new_bound);

					//printf("New bound: %d\n", new_bound);
				}
			}
			else if (qmi->mode==1) {
				/*coverage awareness*/
				//calculate bound current coverage
				double current_coverage=0;
				if (qmi->bound<=qmi->pdf.mean)
					current_coverage=0;
				else
					current_coverage = (1-(qmi->pdf.variance/(qmi->pdf.variance+
										((qmi->bound-qmi->pdf.mean)*
										(qmi->bound-qmi->pdf.mean)))));

				//printf("Current/Specified coverage: %.2f/%.2f\n", current_coverage, qmi->coverage);

				//if current coverage is deviated more than specified, 
				//call handler with the new coverage of the specified bound
				if (fabs(qmi->coverage-current_coverage)>qmi->deviation) {
					qmi->handler(current_coverage);
					//printf("New coverage: %.2f\n", current_coverage);
				}
			}
		} //end if(qos entry not empty and = ec_id)
	} //end for (scan qos table)
}


/**
 * FUNCTION: reply_wait_info
 *
 * GOAL: Send distributed FD result to application
 *
 * PARAM:
 *		wait_info_pos - index of FD result in wait_info table
 */
void reply_wait_info (int wait_info_pos) {
	//before sending to application, verify if we are monitoring (QoS) the 
	//associated event channel
	process_qos_monitoring(wait_info_pos);
	
	//return info through memory mapping
	fi->from = wait_info[wait_info_pos].from;
	fi->seq = wait_info[wait_info_pos].seq;
	fi->ec_id = wait_info[wait_info_pos].ec_id;
	fi->spec = wait_info[wait_info_pos].spec;

	for (int i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		if (wait_info[wait_info_pos].failure_info[i].empty) {
			fi->failure_info[i].empty = TRUE;
			continue;
		}
		
		fi->failure_info[i].empty = FALSE;
		fi->failure_info[i].id = wait_info[wait_info_pos].failure_info[i].id;

		fi->failure_info[i].fail = 
					wait_info[wait_info_pos].failure_info[i].fail;
		fi->failure_info[i].lateness_avail = 
					wait_info[wait_info_pos].failure_info[i].lateness_avail;
		fi->failure_info[i].lateness_degree = 
					wait_info[wait_info_pos].failure_info[i].lateness_degree;
	}
	//sinalize DLL that info is ready
	SetEvent(wait_info_event);
	
	//terminate failure handler if it exists and no failure occured
	if (wait_info[wait_info_pos].task_id!=NULL && !wait_info[wait_info_pos].any_fail)
		TerminateThread(wait_info[wait_info_pos].task_id, 0);
	
	//free wait_info entry
	wait_info[wait_info_pos].failure_at_time = -1;
}	


/**
 * FUNCTION: get_wait_info_pos
 *
 * GOAL: returns the index of the wait_info table entry with the 
 *		 received arguments
 *
 * PARAM:
 *		seq - sequence number
 *		from - sender id
 */
int get_wait_info_pos (int seq, unsigned long from) {
	
	for (int i=0; i<MAX_WAIT_INFO_TABLE_SIZE; i++) {
		if (wait_info[i].failure_at_time!=-1 &&	wait_info[i].seq==seq &&
			wait_info[i].from==from)
			return i;
	}
	return -1;
}


/**
 * FUNCTION: processResponse
 *
 * GOAL: processes a distributed timing failure detection response.
 *		 Called by processRead.
 *
 * PARAMS:
 *		destination - destination id
 *		from_ti - info about sender TCB
 *		seq - sequence number that identifies FD operation
 *		fail - indicates if a failure occured
 *		lateness_avail - indicates if lateness degree is available
 *		lateness_degree - indicates lateness degree, if available
 */
void processResponse (unsigned long destination, tcb_info_t *from_ti, int seq, char fail, 
					  bool lateness_avail, int lateness_degree) {
	
	debug_tcb(("processResponse(from=%u,seq=%d,dest=%u)\n",from_ti->id, seq, destination));
		
	/*
#ifdef LOG_QUEUES

	//if there is a <from,seq-1>, reuse that entry
	bool found=FALSE;
	for (int k=0; k<procResponse_log_size && !found; k++) {
		if (procResponse_log[k].from == from &&
			procResponse_log[k].originalFrom == destination &&
				procResponse_log[k].seq == seq-1) {
			
					found = TRUE;
					procResponse_log[k].seq = seq;
		}
	}	
			
	if (!found) {
		procResponse_log[procResponse_log_size].from = from;
		procResponse_log[procResponse_log_size].originalFrom = destination;
		procResponse_log[procResponse_log_size++].seq = seq;
	}

#endif
   */

	//if it's a reponse to a request sent by this TCB, decrement counter
	if (destination==my_id) {
		/*
		tcb_info_t* ti = getTCB(my_id);
		if (ti==NULL) {
			printf("Error: I doesn't exist!\n");
			return;
		}
		*/
		from_ti->number_of_detections--;
		debug_tcb(("processResponse: Remaining n_detections=%d\n", from_ti->number_of_detections));
	}

	//get position of wait_info table entry that stores FD operation status
	int wait_info_pos = get_wait_info_pos(seq,destination);
	//if entry is not found, we are not expecting this response
	if (wait_info_pos==-1)
		return;

	//if a failed occured and it's the first one received
	if (fail && !wait_info[wait_info_pos].any_fail) {
	
		//if it's a response to a request sent by this TCB, wakeup failure handler
		if (!wait_info[wait_info_pos].remote_fd &&
			wait_info[wait_info_pos].task_id!=NULL)
			
			ResumeThread(wait_info[wait_info_pos].task_id);

		wait_info[wait_info_pos].any_fail = TRUE;
	}

	//update fail info of the machine sending the response
	char index=-1;
	for (int i=0; i<MAX_NUMBER_OF_MACHINES && index==-1; i++) {
		if (!wait_info[wait_info_pos].failure_info[i].empty &&
			 wait_info[wait_info_pos].failure_info[i].id == from_ti->id)
			 index = i;
	}
	if (index==-1) {
		printf("Error: Couldn't find failure_info\n");
		return;
	}
	wait_info[wait_info_pos].failure_info[index].fail = fail;
	wait_info[wait_info_pos].failure_info[index].lateness_avail = lateness_avail;
	if (lateness_avail)
		wait_info[wait_info_pos].failure_info[index].lateness_degree = 
															lateness_degree;
	else
		wait_info[wait_info_pos].failure_info[index].lateness_degree = 0;

	//decrement n_dests
	wait_info[wait_info_pos].n_dests --;
	//if no more expected answers, send info to application
	if(wait_info[wait_info_pos].n_dests==0)
		reply_wait_info(wait_info_pos);
}


/**
 * FUNCTION: processRead
 *
 * GOAL: entry point for the distributed timing failure detection service.
 *		 Called by RecvThread.
 *
 * PARAMS:
 *		dm - data msg
 *		roundtripVal - maximum delivery time
 *		time_of_receipt - reception timestamp
 *		ti - Info about sender TCB
 */
void processRead(data_msg_t *dm, int roundtripVal, int time_of_receipt, 
				 tcb_info_t *ti) {
	
	debug_tcb(("processRead(rtrip=%d,time_rec=%d)\n",roundtripVal, time_of_receipt));
/*
#ifdef LOG_QUEUES

	//if there is a <from,seq-1>, reuse that entry
	bool found=FALSE;
	for (int k=0; k<procRead_log_size && !found; k++) {
		if (procRead_log[k].from == dm->hdr.machine_src_id &&
				procRead_log[k].seq == dm->hdr.seq-1) {
			
					found = TRUE;
					procRead_log[k].seq = dm->hdr.seq;
		}
	}	
			
	if (!found) {
		procRead_log[procRead_log_size].from = dm->hdr.machine_src_id;
		procRead_log[procRead_log_size++].seq = dm->hdr.seq;
	}

#endif
*/

	info_field_t *info_ptr = (info_field_t*)dm->data;

	//scan info_fields and:
	//if a request, call processRequest;
	//if a response, call processResponse
	for (int i=0; i<dm->hdr.num_info_fields; i++) {
		if (info_ptr->type==0)
			processRequest(ti, info_ptr->destination,
						   info_ptr->seq, info_ptr->time_to_failure,
						   roundtripVal, time_of_receipt, info_ptr->remote_fd,
						   info_ptr->ec_id, info_ptr->spec);
		else {
			//only process reponse if one of the following is true:
			//a) response is for me
			//b) response is for all
			//c) reponse was sent by me
			if ((info_ptr->destination[0]==my_id)||(info_ptr->type==2)||
				(dm->hdr.machine_src_id==my_id)) {
				processResponse(info_ptr->destination[0], ti, 
								info_ptr->seq, info_ptr->fail,
								info_ptr->lateness_avail, info_ptr->lateness_degree);
			}
			/*
			else {
				printf("Warning: Discarding msg on processRead: src=%d, seq=%d\n",
						dm->hdr.machine_src_id, info_ptr->seq);	
			}
			*/
		}
		info_ptr++;
	}
}


/**
 * FUNCTION: detect_crashes
 *
 * GOAL: To verify if any TCB has crashed
 */
void detect_crashes() {
	
	int last_life_signal;

	for (int i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		if (!membership[i].empty && membership[i].id != my_id) {
			//when no data msg was yet received
			if (membership[i].last_rcv_info.timestamp==0) {
				last_life_signal = membership[i].arrival;
			}
			//when a data msg was already received
			else
				last_life_signal = membership[i].last_rcv_info.timestamp;
			
			if (last_life_signal < GetTickCount()-(SEND_PERIOD+RECV_PERIOD+SECURITY_MARGIN_FOR_NODE_CRASH_DETECTION+MAXIMUM_DELIVERY_DELAY-MINIMUM_DELIVERY_DELAY)) {
				//Delete TCB from membership
				int ret = deleteTCB(membership[i].id);
				if (ret!=0) {
					printf("Error: Couldn't delete TCB %u because it doesn't exist\n", membership[i].id);
					continue;
				}
				//print IP of deleted TCB
				struct in_addr in;
				in.S_un.S_addr = membership[i].id;
				printf("TCB %s has left!\n", inet_ntoa(in));
			}
		}
	}
}


/**
 * FUNCTION: check_blocked_tcb
 *
 * GOAL: To verify if there are pending distributed failure detections 
 *		 due to crashed TCBs.
 */
void check_blocked_tfd() {
	
	for (int i=0; i<MAX_WAIT_INFO_TABLE_SIZE; i++) {
		//suspect only of late detections
		if (wait_info[i].failure_at_time != -1 && 
			(GetTickCount()>=(wait_info[i].failure_at_time+SEND_PERIOD+RECV_PERIOD))) {
		
			//indicates if detection result can be delivered to application
			bool can_deliver=TRUE;

			for (int j=0; j<MAX_NUMBER_OF_MACHINES; j++) {
				//if a destination has crashed
				if (wait_info[i].destination[j]!=0 &&
					getTCB(wait_info[i].destination[j])==NULL) {
					
					//and if we did not yet receive an answer, we will never receive
					if (wait_info[i].failure_info[j].fail==2) {
						wait_info[i].failure_info[j].fail = 1;
						wait_info[i].n_dests--;
						wait_info[i].failure_info[j].lateness_avail = FALSE;
						//execute failure handler if not already executed
						if (!wait_info[i].any_fail) {
							if (!wait_info[i].remote_fd &&
								 wait_info[i].task_id!=NULL)
									ResumeThread(wait_info[i].task_id);

							wait_info[i].any_fail = TRUE;
						}
					}
				}
				//if the destination has not crashed and we may yet receive a response
				else if (wait_info[i].destination[j]!=0 &&
					     wait_info[i].failure_info[j].fail==2) {
					
						can_deliver = FALSE;
				}
			} //end for MACHINES
			
			if (can_deliver) {
				//printf("check_blocked_tfd: replying wait_info (from=%u, seq=%d)\n", 
				//				wait_info[i].from, wait_info[i].seq);

				reply_wait_info(i);
			}

		} //end if late detection
	} //end for wait_info
}



/**
 * THREAD: RecvThread
 *
 * GOAL: to periodically receive and process a msg from other TCB
 */
DWORD WINAPI RecvThread (LPVOID param) {
	//printf("RecvThread running\n");
	int n_seq, *last_seq_num;
	unsigned long id;
	tcb_info_t* ti;

	while(1) {
		data_msg_t dm;
		char buffer[sizeof(data_msg_t)];
		int nbytes;

/*
		EnterCriticalSection(&loopback_critical_section);
		//check if at least one msg was sent in loopback
		if (loop_first!=-1 && loop_last!=-1) {
			//return first msg in buffer
			memcpy(buffer, loopbt[loop_first].buffer, loopbt[loop_first].size);
			nbytes = loopbt[loop_first].size;
			//printf("Removed msg from loop pos %d\n", loop_first);
			//if buffer becomes empty
			if (loop_first == loop_last) {
				loop_first = loop_last = -1;
			}
			else loop_first = (loop_first+1)%MAX_SIZE_LOOPBACK_BUFFER;
			printf("Msg from loopback\n");
			LeaveCriticalSection(&loopback_critical_section);
		}
		//if not, receive from network
		else {
			LeaveCriticalSection(&loopback_critical_section);
			*/
			//receive msg
			struct sockaddr_in source;
			int source_size = sizeof(source);
			nbytes=recvfrom(s, buffer, sizeof(data_msg_t), NULL, 
							(struct sockaddr*)&source, &source_size);
		/*}*/
			
		//for measuring the time it takes a cycle
		int now = GetTickCount();
	
		//recved a id msg
		if (buffer[0]==ID) {
			//verify if size is correct
			if (nbytes!=sizeof(id_msg_t)) {
				printf("Error: ID msg with incorrect size\n");
				continue;
			}
			else {
				id_msg_t imt;
				memcpy((char*)&imt, buffer, nbytes);
				//verify if id is already in our membership
				tcb_info_t* ti = getTCB(imt.machine_src_id);
				if (ti==NULL) {
					//Add new TCB
					int ret = addTCB(imt.machine_src_id);
					if (ret!=0) {
						//calculate IP of TCB 
						struct in_addr in;
						in.S_un.S_addr = imt.machine_src_id;
						printf("Error: Couldn't add new TCB %s because the maximum number of TCBs are already running\n", inet_ntoa(in));
						continue;
					}
					//print IP of new TCB
					struct in_addr in;
					in.S_un.S_addr = imt.machine_src_id;
					printf("TCB %s has joined!\n", inet_ntoa(in));
				}
				continue;
			}
		}
		//recved a data msg
		else if (buffer[0]==DATA) {
			//verify if size is correct
			if (nbytes<sizeof(data_msg_hdr_t)) {
				printf("Error: Data msg with incorrect size\n");
				continue;
			}
			else {
				memcpy((char*)&dm, buffer, nbytes);
				n_seq=dm.hdr.seq;
				id=dm.hdr.machine_src_id;
				ti = getTCB(id);
				if (ti==NULL) {
					//Add new TCB
					int ret = addTCB(id);
					if (ret!=0) {
						//calculate IP of TCB 
						struct in_addr in;
						in.S_un.S_addr = id;
						printf("Error: Couldn't add new TCB %s because the maximum number of TCBs are already running\n", inet_ntoa(in));
						continue;
					}
					ti = getTCB(id);
					if (ti==NULL) {
						printf("Error: Couldn't find TCB that was just detected\n");
						continue;
					}
					//print IP of new TCB
					struct in_addr in;
					in.S_un.S_addr = id;
					printf("TCB %s has joined!\n", inet_ntoa(in));
				}
				last_seq_num=&(ti->last_rcv_info.seq);
				debug_net(("Data seq %d from %d\n", n_seq, id));
			}
		}
		//received an unknown message type
		else {
			printf("FATAL Error: Message type is unknown: %d\n", buffer[0]);
			continue;
		}
		//debug_net(("Recved %d bytes from %d\n", nbytes, m.id));
	
		/* check if msg should be processed */
		//if it's the first one, init last_seq_num
		if (*last_seq_num==-1) {
			*last_seq_num=n_seq-1;
			ti->first_seq=n_seq;
		}
		//if msg sequence number is less than the last received,
		//it's an error
		if (n_seq<*last_seq_num) {
			printf("FATAL Error: msg seq num<last seq\n");
			return -1;
		}
		//if msg sequence number is equal to the last received,
		//it's a duplicate: discard it!
		else if (n_seq==*last_seq_num) {
			debug_net(("Discarding duplicate\n"));
		}
		else {
			//check if there were omissions
			if (n_seq>(*last_seq_num)+1) {
				
				//printf("FATAL error: %d omissions from %s, seq=%d\n", 
				//				n_seq-((*last_seq_num)+1), inet_ntoa(in), n_seq);

				//don't consider omissions from myself
				if (id!=my_id) {
				 //calculate and print IP of TCB 
				 struct in_addr in;
				 in.S_un.S_addr = id;

				 printf("Warning: %d omissions from %s, seq=%d\n", 
								n_seq-((*last_seq_num)+1), inet_ntoa(in), n_seq);

				 n_omissions+=n_seq-((*last_seq_num)+1);
				}
				
				/*
				//don't delete myself!
				if (id!=my_id) {
				 EnterCriticalSection(&tcb_critical_section);

				 //return -1;
				 //Delete sender TCB from membership
				 int ret = deleteTCB(id);
				 if (ret!=0) {
					printf("Error: Couldn't delete TCB %u because it doesn't exist\n", id);
					LeaveCriticalSection(&tcb_critical_section);
					continue;
				 }
				 //print IP of deleted TCB
				 in.S_un.S_addr = id;
				 printf("TCB %s has left!\n", inet_ntoa(in));
				 LeaveCriticalSection(&tcb_critical_section);
				 continue;
				}
				*/
			}
		

			EnterCriticalSection(&tcb_critical_section);
		
			//update last_rcv_info (synchronize with SendThread)
			int timestamp = GetTickCount();
			//update last_rcv_info
			ti->last_rcv_info.timestamp = timestamp;
			ti->last_rcv_info.seq = dm.hdr.seq;
			ti->last_rcv_info.snd_time = dm.hdr.snd_time;
			
			//update source ec_id
			ti->ec_ids = dm.hdr.ec_id;
		
			debug_net(("Data recved from %u, seq=%d\n", 
					dm.hdr.machine_src_id, dm.hdr.seq));
			
			//calculate maximum delivery time
			int roundtripVal = checkRoundTrip(&dm,timestamp);
			//if error, exit
			if (roundtripVal > MAXIMUM_DELIVERY_DELAY) {
				/*
				//don't delete myself
				if (id!=my_id) {
				 //Delete sender TCB from membership
				 int ret = deleteTCB(id);
				 if (ret!=0) {
					printf("Error: Couldn't delete TCB %u because it doesn't exist\n", id);
					LeaveCriticalSection(&tcb_critical_section);
					continue;
				 }
				 //print IP of deleted TCB
				 struct in_addr in;
				 in.S_un.S_addr = id;
				 printf("TCB %s has left!\n", inet_ntoa(in));
				 LeaveCriticalSection(&tcb_critical_section);
				}
				continue;
				//return -1;
				*/
			}
			
			//update global clock
			if ((n_seq - ti->first_seq)<10) {
				//for first 5 msgs, call update_vclock in init_state
				update_vclock(ti, dm.hdr.vclock,
								  roundtripVal, timestamp, true);
			}
			else {
				//for other msgs, call update_vclock normally
				update_vclock(ti, dm.hdr.vclock,
								  roundtripVal, timestamp, false);
			}


			//crash failure detector
			detect_crashes();

			//if (roundtripVal>-1) {
			//if we received distributed FD data
			if (dm.hdr.num_info_fields>0) {
				//pass msg to distributed timing failure detection service
				processRead(&dm, roundtripVal, timestamp, ti);
			}
			
			//check for timing failure detections pending due to crashed TCBs
			check_blocked_tfd();

			LeaveCriticalSection(&tcb_critical_section);
				
#ifdef TCB_MEASUREMENTS
			//measure only the values different from the theoretical maximum
			if (roundtripVal!=MAXIMUM_DELIVERY_DELAY) {
				/* measures the (average/min/max) values of roundtripval */
				tcb_info_t* my_ti = getTCB(dm.hdr.machine_src_id);
				if (my_ti == NULL) {
					printf("Error: Couldn't find myself!\n");
					return -1;
				}
				//update the sum of all roundtrip values
				my_ti->total_roundtripVal+= roundtripVal;
				//update the number of roundtrip values
				my_ti->num_roundtripVal++;
				//check if it's the maximum
				if (roundtripVal>my_ti->max_roundtripVal)
					my_ti->max_roundtripVal=roundtripVal;
				//check if it's the minimum
				if (roundtripVal<my_ti->min_roundtripVal)
					my_ti->min_roundtripVal=roundtripVal;
			}
			/*
			/*print values necessary for TFDmax calculation
			if (num_roundtripVal[id]!=0 && (num_roundtripVal[id]%100000)==0) {
				//prints average roundtrip for each 100.000 roundtrip values recved
				printf("%d msg,(AV/MIN/MAX)rtrip from %d=(%.2f/%d/%d) %d om.\n",
						num_roundtripVal[id], id, 
						((double)total_roundtripVal[id])/num_roundtripVal[id],
						min_roundtripVal[id],max_roundtripVal[id], n_omissions);
				
				//prints maximum time that a msg waits in the send queue and in the 
				//receive queue
				printf("send_q latency=(%.2f/%d/%d), rcv_q latency=%d,num_rcv_queue_overflow=%d\n", 
							((double)total_in_send_queue_latency)/num_in_send_queue_latency,
							min_in_send_queue_latency,
							max_in_send_queue_latency, 
							max_before_recv_queue_latency,
							num_rcv_queue_overflow);
				
				//prints average time that a token takes to circle by all TCBs
				printf("%d tokens,(AV/MIN/MAX)token delay:(%.2f/%d/%d)\n",
						num_token_delay, 
						((double)total_token_delay)/num_token_delay,
						min_token_delay,max_token_delay);

			}
			*/
#endif		
		}
		//update last sequence number received
		*last_seq_num=n_seq;

		int elapsed = GetTickCount() - now;
		if (RECV_PERIOD-elapsed>0)
			Sleep(RECV_PERIOD-elapsed);
		else 
			printf("Warning: RecvThread runned during %d ms and RECV_PERIOD=%d\n",
								elapsed, RECV_PERIOD);
		

	}
	//never reaches here	
	//return 0;
}





/**
 * FUNCTION init_variables
 *
 * GOAL: Initialize variables
 */
int init_variables() {

	//init membership
	for (int i=MAX_NUMBER_OF_MACHINES;i--;)
		membership[i].empty=TRUE;
		
	//init local measures vector
	for (i=MAX_LOCAL_MEASURES;i--;)		
	    localMeasures[i]=0;

	//init timely executions table
	for (i=MAX_EXEC;i--;)		
	    executions[i].func=NULL;

	//init local timing failure detection table
	for (i=MAX_LOCAL_FD;i--;)		
	    localFD[i].func=NULL;
		
	//init number of distributed FD detections running in each send period
	n_sends_in_this_slot=0;

	//init send_info table
	for (i=MAX_SEND_INFO_TABLE_SIZE; i--;) {
		send_info[i].time_of_generation=-1;
	}

	//init receive_info table
	for (i=MAX_RCV_INFO_TABLE_SIZE; i--;) {
		receive_info[i].failure_at_time=-1;
	}

	//init receive_payload_info table
	for (i=MAX_RCV_PLOAD_TABLE_SIZE; i--;) {
		receive_payload_info[i].time_of_receipt=-1;
	}

	//init wait_info table
	for (i=MAX_WAIT_INFO_TABLE_SIZE; i--;) {
		wait_info[i].failure_at_time=-1;
	}

	//init qos_monitor_info table
	for (i=MAX_QOS_MONITOR_INFO_TABLE_SIZE; i--;) {
		qos_monitor_info[i].empty = TRUE;
	}

	init_vclock();
	return 0;
}


/**
 * FUNCTION print_info
 *
 * GOAL: Print TCB info
 */
void print_info() {
	//debug_net(("Id:%d, Next:%d, Previous:%d\n",my_id,successor,predecessor));

	//printf("Data Omission Degree: %d\n", DATA_OMISSION_DEGREE);
	//printf("Max. packet size: %d\n", MAX_PACKET_SIZE);
	//printf("Sizeof data_msg: %d\n", sizeof(data_msg));
	//printf("Max. info fields: %d\n", MAX_NUM_INFO_FIELDS);
	
	printf("Max. dist. FD services: %d\n", MAX_DISTRIBUTED_FD_SERVICES);
	//printf("Send period: %d\n", SEND_PERIOD);
	//printf("Recv period: %d\n", RECV_PERIOD);
	//printf("Crash latency: %d ms\n", CRASH_LATENCY);
	printf("TFDmax with sender FD: %d ms\n", TFDMAX_WITH_SENDER_FD);
	printf("TFDmax with remote FD: %d ms\n", TFDMAX_WITH_REMOTE_FD);
}


/**
 * FUNCTION init_id
 *
 * GOAL: To initialize our id
 */
int init_id() {
	
	//randomly generate our id 
	//srand(GetTickCount());
	//my_id = rand()%256;

	char ip[16];
	//read IP from configuration file
	FILE* f = fopen("/Windows/Start Menu/tcb_config.txt", "r");
	if (f==NULL){
		printf("Error: Config file not found\n");
		Sleep(5000);
		return -1;
	}
	if(fscanf(f, "%s", ip)!=1){
		printf("Error: Could not read IP from config file\n");
		Sleep(5000);
		return -1;
	}
	fclose(f);
	my_id = inet_addr(ip);
	if (my_id==INADDR_NONE) {
		printf("Error: Could not read IP from config file\n");
		Sleep(5000);
		return -1;
	}	
	//print IP of TCB 
	struct in_addr in;
	in.S_un.S_addr = my_id;
	printf("TCB id: %s\n", inet_ntoa(in));

	return 0;
}



/**
 * FUNCTION init_network
 *
 * GOAL: Create and bind socket
 *
 * RETURNS: 0 if OK, -1 if error
 */
int init_network() {

	TCBstartTime = GetTickCount();

	//create socket for communication
	s = socket(AF_INET, SOCK_DGRAM, NULL);
	if (s==INVALID_SOCKET) {
		_wperror(L"Error creating TCB socket");
		Sleep(5000);
		return -1;
	}


#ifdef REAL_BROADCAST
	//enable broadcast
	int optval=1;
	if (setsockopt(s, SOL_SOCKET, SO_BROADCAST, (char*)&optval, sizeof(int))!=0) {
		_wperror(L"Error enabling broadcast");
		Sleep(5000);
		return -1;
	}
	printf("Real-Broadcast enabled!\n");
#endif
	

	//bind socket
	struct sockaddr_in sa;
	sa.sin_family=AF_INET;
	sa.sin_addr.s_addr= my_id;
	sa.sin_port=htons(TCB_PORT);
	if (bind(s, (struct sockaddr*)&sa, sizeof(sa))!=0) {
		_wperror(L"Error binding TCB socket");
		Sleep(5000);
		return -1;
	}
	

	//debug_net(("Socket created and binded\n"));
#ifdef REAL_BROADCAST
	debug_net(("Using real broadcast\n"));
#endif

#ifdef EMUL_BROADCAST
	debug_net(("Using emulated broadcast\n"));
#endif
	return 0;
}


/**
 * FUNCTION: print_exec_info
 *
 * GOAL: Prints execution info when TCB is shutting down
 */
void print_exec_info () {
	int i;
	int TCBtimeElapsed = GetTickCount() - TCBstartTime;
	printf("TCB runned %d seconds\n", TCBtimeElapsed/1000);


#ifdef TCB_MEASUREMENTS
	for (i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		if (!membership[i].empty) {
		 //get IP of TCB to prin it below
		 struct in_addr in;
		 in.S_un.S_addr = membership[i].id;
		 
		 //prints average roundtrip
		 printf("%d msg,(AV/MIN/MAX)delivery delay from %s=(%.2f/%d/%d).\n",
					membership[i].num_roundtripVal, inet_ntoa(in), 
					((double)membership[i].total_roundtripVal)/membership[i].num_roundtripVal,
						membership[i].min_roundtripVal,membership[i].max_roundtripVal);
		}
	}
#endif	
	
	/*
	//prints number of data msg that were lost
	printf("#omissions=%d\n", n_omissions);
	printf("Last data sent=%d\n", data_seq_num);
	for (i=0; i<MAX_NUMBER_OF_MACHINES; i++) {
		if (!membership[i].empty)
		  printf("Last data recved from %d=%d\n",
						membership[i].id,membership[i].last_rcv_info.seq);
	}
	*/
}




/**
 * Initializes TCB, launches threads, ...
 */
int main(int argc, char* argv[]) {

	//prints TCB version
	printf("TCB version: %d.%d.%d\n", MAJOR_VERSION, MINOR_VERSION, TINY_VERSION);

	int ret = init_id();
	if (ret!=0)
		return ret;

	print_info();

	ret = init_variables();
	if (ret!=0)
		return ret;

	//add TCB to membership
	ret = addTCB(my_id);
	if (ret!=0)
		return ret;

	ret = init_network();
	if (ret!=0)
		return ret;
	
	//init tcb synchronization
	InitializeCriticalSection(&tcb_critical_section);

	//init loopback buffer synchronization
	InitializeCriticalSection(&loopback_critical_section);

	/*init tcb services auxiliary variables*/
	//create event to sinalize dll that an answer is ready
	dll_wait_event = CreateEvent(NULL, TRUE, FALSE, L"dll_wait_event");
	if (dll_wait_event==NULL) {
		_wperror(L"Error creating dll wait event\n");
		wait_user();
		return -1;
	}
	//create event to dll sinalize us that a request is ready
	tcb_wait_event = CreateEvent(NULL, TRUE, FALSE, L"tcb_wait_event");
	if (tcb_wait_event==NULL) {
		_wperror(L"Error creating tcb wait event\n");
		wait_user();
		return -1;
	}
	//create event sinalize dll to ... 
	wait_info_event = CreateEvent(NULL, TRUE, FALSE, L"wait_info_event");
	if (wait_info_event==NULL) {
		_wperror(L"Error creating wait info event\n");
		wait_user();
		return -1;
	}
	//create a memory mapping to be used in TCB DLL - ServicesThread comm
	HANDLE fmap = CreateFileMapping(INVALID_HANDLE_VALUE, 
		NULL, PAGE_READWRITE, 0, sizeof(service_request_response_t), 
		L"tcb_memory_map");
	if (fmap==NULL) {
		_wperror(L"Error creating memory mapping\n");
		wait_user();
		return -1;
	}
	//associate the service_request_response with the memory mapping
	srrt =(service_request_response_t *) 
			MapViewOfFile(fmap, FILE_MAP_WRITE, 0, 0, 0);
	if (srrt==NULL) {
		_wperror(L"Error mapping memory\n");
		wait_user();
		return -1;
	}
	//fill with zeros
	memset(srrt, 0, sizeof(service_request_response_t));

	//create a memory mapping to be used in DistFDService->TCB DLL waitInfo() comm
	HANDLE fmap1 = CreateFileMapping(INVALID_HANDLE_VALUE, 
		NULL, PAGE_READWRITE, 0, sizeof(fail_info_to_app_t), 
		L"wait_info_memory_map");
	if (fmap1==NULL) {
		_wperror(L"Error creating memory mapping 2\n");
		wait_user();
		return -1;
	}
	//associate the fail_info_to_app_t with the memory mapping
	fi =(fail_info_to_app_t *) 
			MapViewOfFile(fmap1, FILE_MAP_WRITE, 0, 0, 0);
	if (fi==NULL) {
		_wperror(L"Error mapping memory 2\n");
		wait_user();
		return -1;
	}
	//fill with zeros
	memset(fi, 0, sizeof(fail_info_to_app_t));

	//create send task (sends tcb messages)
	hSend = CreateThread(NULL,0,SendThread,NULL,CREATE_SUSPENDED,NULL);
	
	//create receive task (receives tcb messages)
	hRecv = CreateThread(NULL,0,RecvThread,NULL,CREATE_SUSPENDED,NULL);

	//create services task 
	//(receives services requests from TCB DLL and answers them)
	HANDLE hServ = CreateThread(NULL,0,ServicesThread,NULL,CREATE_SUSPENDED,NULL);

	if (!hSend || !hRecv || !hServ){
		printf("Error creating threads!\n");
		DeleteCriticalSection(&tcb_critical_section);
		DeleteCriticalSection(&loopback_critical_section);
		//unmap shared memory
		if(!UnmapViewOfFile(srrt)){
			_wperror(L"Error unmaping view of file\n");
			wait_user();
			return -1;
		}
		//close memory map handle
		CloseHandle(fmap);
		//close event handles
		CloseHandle(tcb_wait_event);
		CloseHandle(dll_wait_event);
		wait_user();
		return -1;
	}

	//maximum priority for send and receive thread
	CeSetThreadPriority(hSend, 0);
	CeSetThreadPriority(hRecv, 0);
	//maximum-1 priority for services thread
	CeSetThreadPriority(hServ, 1);
	

	//start receive task
	ResumeThread(hRecv);
	debug_tcb(("Recv Thread started\n"));

	//start send task
	ResumeThread(hSend);
	debug_tcb(("Send Thread started\n"));
	
	//start services task
	ResumeThread(hServ);
	debug_tcb_services(("Services Thread started\n"));

	printf("TCB started\n");

	//wait until recv thread returns
	WaitForSingleObject(hRecv, INFINITE);
	
	/*shutdown TCB*/
	printf("Shutting down TCB\n");
	debug_tcb(("Terminating Send thread\n"));
	TerminateThread(hSend, 0);
	debug_tcb(("Terminating Recv thread\n"));
	TerminateThread(hRecv, 0);
	debug_tcb_services(("Terminating Services thread\n"));
	TerminateThread(hServ, 0);
	
	
	CloseHandle(hSend);
	CloseHandle(hRecv);
	CloseHandle(hServ);

	//unmap shared memory
	if(!UnmapViewOfFile(srrt)){
		_wperror(L"Error unmaping view of file\n");
		wait_user();
		return -1;
	}
	//close memory map handle
	CloseHandle(fmap);
	//close event handles
	CloseHandle(tcb_wait_event);
	CloseHandle(dll_wait_event);
	
	DeleteCriticalSection(&tcb_critical_section);
	DeleteCriticalSection(&loopback_critical_section);

	print_exec_info();

	//wait for user input
	wait_user();
	return 0;
}
