/**
 * TCB for Windows CE 3.0
 *
 * This header file defines the TCB configuration parameters.
 *
 * @author Paulo Sousa
 * (based on RTAI TCB implementation by Pedro Martins)
 * @date 18/Nov/2002
 */

#ifndef CONFIG_PARAMETERS_TCB
#define CONFIG_PARAMETERS_TCB

//our port (change if port is being used by other application)
#define TCB_PORT 25000

//turn on/off TCB measurements
//#define TCB_MEASUREMENTS

//Maximum number of machines running TCB
#define MAX_NUMBER_OF_MACHINES 3

//Data messages omission degree
#define DATA_OMISSION_DEGREE 0

//802.11 MTU 
#define MTU 2000

//Size of socket receive buffer
#define SOCK_RCV_BUF_SIZE 4096

//local clock drift rate
#define CLOCK_DRIFT_RATE 0.000005


//Maximum packet size according to MTU,SOCK_RCV_BUF_SIZE,
//NUMBER_OF_MACHINES and DATA_OMISSION_DEGREE
#if MTU<(SOCK_RCV_BUF_SIZE/MAX_NUMBER_OF_MACHINES)

 #define MAX_PACKET_SIZE MTU/(DATA_OMISSION_DEGREE+1)

#else

 #define MAX_PACKET_SIZE (SOCK_RCV_BUF_SIZE/MAX_NUMBER_OF_MACHINES)/(DATA_OMISSION_DEGREE+1)

#endif



//#define MAX_PACKET_SIZE 500



/*Type of Broadcast*/

//for real broadcast, define the broadcast address to use
//#define REAL_BROADCAST "169.254.0.255"

//emulated broadcast= send unicast packet to all machines and 
//broadcast an id msg to all
#define EMUL_BROADCAST "169.254.0.255"

//Period of TCB SendThread (milliseconds)
#define SEND_PERIOD 400

//Period of TCB RecvThread (milliseconds)
#define RECV_PERIOD 50

//To guarantee that only failed TCB are detected as crashed
#define SECURITY_MARGIN_FOR_NODE_CRASH_DETECTION  500

//maximum number of samples that can be considered to construct a PDF
#define MAX_NUM_SAMPLES 100

//maximum number of simultaneous local measures
#define MAX_LOCAL_MEASURES 10

//maximum number of simultaneous timely executions
#define MAX_EXEC 10

//maximum number of simultaneous local timing failure detections
#define MAX_LOCAL_FD 10

//minimum delivery time of a data msg (milliseconds)
#define MINIMUM_DELIVERY_DELAY 0

//maximum delivery time of a data msg (milliseconds)
#define MAXIMUM_DELIVERY_DELAY 300

//latency of crash failure detector
#define CRASH_LATENCY (SEND_PERIOD+RECV_PERIOD+SECURITY_MARGIN_FOR_NODE_CRASH_DETECTION+2*MAXIMUM_DELIVERY_DELAY-MINIMUM_DELIVERY_DELAY)

//TFDmax for distributed timing failure detection with FD in sender
#define TFDMAX_WITH_SENDER_FD (((SEND_PERIOD + MAXIMUM_DELIVERY_DELAY + RECV_PERIOD)*2)+CRASH_LATENCY)

//TFDmax for distributed timing failure detection with FD in recipients
#define TFDMAX_WITH_REMOTE_FD (SEND_PERIOD + MAXIMUM_DELIVERY_DELAY + RECV_PERIOD)

#endif
